Require Import List.
Require Import Arith.
Require Import Program.Equality.
Require Import Omega.
Require Import NPeano.

Hint Extern 2 =>
  match goal with
  | [ |- _ <= _ ] => omega
  | [ |- _ < _ ] => omega
  end.

Hint Extern 2 =>
  match goal with
  | [ H : context[?n + 1] |- _ ] => 
    rewrite (plus_comm n 1) in H; simpl in H
  | [ |- context[?n + 1] ] => 
    rewrite (plus_comm n 1); simpl
  end.


Theorem plus_comm' : forall n m, m + n = n + m.
Proof. intros; rewrite plus_comm; auto. Qed.

Theorem plus_n_Sm' : forall n m, n + S m = (S n) + m.
Proof. intros; omega.  Qed.
Ltac swap_n_Sm := rewrite plus_n_Sm'.

Theorem pred_plus_n_Sm : forall n m, pred (n + S m) = n + m.
Proof.  intros; omega.  Qed.
Hint Rewrite pred_plus_n_Sm : step.


Hint Rewrite (plus_comm' 1) : step.
Hint Rewrite (plus_comm' 2) : step.
Hint Rewrite (plus_comm' 3) : step.

Ltac rewrite_ih :=
  repeat (match goal with
  | [ H : _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  end).

(* This file contains basic facts about the substitution and lift
   operations on terms.  It's all De Bruijn index wrangling *)
Ltac compares_cleanup :=
    match goal with
    | [ H : nat_compare _ _ = Eq |- _ ] => apply nat_compare_eq in H
    | [ H : nat_compare _ _ = Lt |- _ ] => apply nat_compare_lt in H
    | [ H : nat_compare _ _ = Gt |- _ ] => apply nat_compare_gt in H
    end.
  
Ltac destruct_compares n m :=
  remember (nat_compare n m);
  match goal with | [ H : ?comp = nat_compare n m |- _ ] =>
    apply eq_sym in H;
    destruct comp
  end;
  compares_cleanup.
  

Ltac clarify_compares :=
  match goal with
  | [ |- context[match nat_compare ?n ?m with
                 | Lt => _ | Eq => _ | Gt => _ 
                 end]
    ] => destruct_compares n m
  | [ H : ?P (match nat_compare ?n ?m with
              | Lt => _ | Eq => _ | Gt => _ 
              end) _ |- _
    ] => destruct_compares n m
  | [ H : ?P _ (match nat_compare ?n ?m with
                | Lt => _ | Eq => _ | Gt => _ 
                end) |- _
    ] => destruct_compares n m
  | [ H : context[(match nat_compare ?n ?m with
                | Lt => _ | Eq => _ | Gt => _ 
                end)] |- _
    ] => destruct_compares n m
  end.

Ltac db_var_case := 
  repeat clarify_compares; simpl in *; intuition;
  repeat clarify_compares; simpl in *; autorewrite with step in *; intuition.

Ltac invert_compare :=
  match goal with
  | [ H : match nat_compare _ _ with | Eq => _ | Lt => _ | Gt => _ end = _
     |- _ ] => clarify_compares; inversion H
  end.

Ltac clarify_max :=
  repeat (match goal with
  | [ H : max _ _ <= _ |- _ ] =>
    apply Nat.max_lub_iff in H; destruct H
  | [ H : max _ _ < _ |- _ ] =>
    apply Nat.max_lub_lt_iff in H; destruct H
  end).


Ltac inst_equalities := 
  repeat (match goal with
  | [ H : forall _ _, _ ++ _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ ++ _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, (_ ++ _ :: _) ++ _ :: _ = _ -> _ |- _ ] =>
    try rewrite <- app_assoc in H; try rewrite <- app_comm_cons in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H

  | [ H : forall _ _ _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
(*
  | [ H : forall G, ctx_wf (G ++ ?G3) -> _,
      H2 : ctx_wf (?G2 ++ ?G3)
     |- context[lift_c _ (length ?G2) _] ] => let Hwf := fresh "Hwf" in 
    assert (Hwf := H _ H2); clear H; rename Hwf into H 
*)
  | [ H : forall _ _ _ _, _ ++ _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, (_ ++ _ :: _) ++ _ :: _ = _ -> _ |- _ ] =>
    try rewrite <- app_assoc in H; try rewrite <- app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H

(*

  | [ H : forall _ _ _ _ _ _ _, _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ ++ _ :: ?G = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing ?G _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H

  | [ H : forall _ _ _ _ _ _ _, _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ :: _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H

  | [ H : forall _ _ _ _ _ _ _, _ ++ _ :: _ ++ (_,({ _ ~> _ } _)) :: _
                              = _ ++ (_,({_ ~> _}_)) :: _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, (_ ++ (_,({_ ~> _}_)) :: _) ++ _
                              = _ ++ (_,({_ ~> _}_)) :: _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    rewrite <- app_assoc in H; rewrite <- app_comm_cons in H;
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ :: _ :: _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
*)
  end).


Ltac destruct_things :=
  repeat (match goal with
  | [ H : exists _, _ |- _ ] => destruct H
  | [ H : _ /\ _ |- _ ] => destruct H
  | [ H : _ \/ _ |- _ ] => destruct H
  end).


Hint Extern 2 =>
  match goal with
  | [ |- _ <= _ ] => omega
  | [ |- _ < _ ] => omega
  end.

Hint Extern 2 =>
  match goal with
  | [ H : context[?n + 1] |- _ ] => 
    rewrite (plus_comm n 1) in H; simpl in H
  | [ |- context[?n + 1] ] => 
    rewrite (plus_comm n 1); simpl
  end.


(* Some ltac that should probably be moved *)
Ltac destruct_ih :=
  match goal with
  [ IH : forall _ _ _, _ = _ -> _ = _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ _ eq_refl eq_refl); clear IH
  | [ IH : forall _ _, ?P _ _ = ?P _ _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ eq_refl); clear IH
  | [ IH : forall _ _ _ _, ?P _ = ?P _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ _ _ eq_refl); clear IH
  | [ IH : forall _ _, ?P _ = ?P _ -> ?Q _ = ?Q _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ eq_refl eq_refl); clear IH
  | [ IH : forall _, ?P _ = ?P _ -> ?A = ?A -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ eq_refl eq_refl); clear IH
  | [ IH : forall _, ?P _ = ?P _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ eq_refl); clear IH
  | [ IH : ?P = ?P -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH eq_refl); clear IH
  | [ IH : forall _ _ _, ?P _ _ _ = ?P _ _ _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ _ eq_refl); clear IH
  end; destruct_things.
