Require Import Arith.
Require Import Bool.
Require Import List.

Require Import shared_tactics.
Require Import shared_defs.
Require Import step_defs.
Require Import step_theory.

Inductive pre_ctx_wf : ctx -> Prop :=
 | PCWFNil  : pre_ctx_wf nil
 | PCWFTrm : forall t A G,
       pretyping G th_log A exp_star
    -> pre_ctx_wf G
    -> pre_ctx_wf ((t,A) :: G)

with pretyping : ctx -> theta -> exp -> exp -> Prop :=
 | PTVar : forall (G1 G3:ctx) (t:theta) (n:nat) (A:exp),
     pre_ctx_wf (G1 ++ (t,A) :: G3) ->
     n = length G1 ->
     pretyping (G1 ++ (t,A) :: G3) th_log (lift A (S n) 0) exp_star ->
     pretyping (G1 ++ (t,A) :: G3) t (exp_var n) (lift A (S n) 0)
 | PTUnboxVal : forall G t v A t',
       pretyping G t  v (exp_at A t')
    -> val v
    (* -> pretyping G th_log A exp_star *)
    -> pretyping G t' v A
 | PTUnboxL : forall G a A t',
       pretyping G th_log a (exp_at A t')
    (* -> pretyping G th_log A exp_star *)
    -> pretyping G t' a A
 | PTUnboxP : forall G a A t',
       pretyping G th_prog a (exp_at A t')
    (* -> pretyping G th_log A exp_star *)
    -> pretyping G th_prog a A
 | PTArr : forall G A B,
       pretyping G th_log A exp_star
    -> FO A
       (* It may look strange that B can assume x to be logical,
          but since A is FO it doesn't matter. *)
    -> pretyping ((th_log,A) :: G) th_log B exp_star
    -> pretyping G th_log (exp_arr A B) exp_star
 | PTApp : forall (G:ctx) (t:theta) (a b A B:exp),
       pretyping G t a (exp_arr A B)
    -> pretyping G t b A
    -> pretyping G th_log ({0 ~> b} B) exp_star
    -> pretyping G t (exp_app a b) ({0 ~> b} B)
 | PTLam : forall (G:ctx)  t(b A B:exp),
       pretyping ((t,A) :: G) t b B
(* Todo: get rid of this premise *)
    -> pretyping G th_log (exp_arr A B) exp_star
    -> pretyping G t (exp_lam b) (exp_arr A B)
(*
 | PTLam : forall (G:ctx) (b A B:exp),
       pretyping ((th_log,A) :: G) th_log b B
    -> pretyping G th_log (exp_arr A B) exp_star
    -> pretyping G th_log (exp_lam b) (exp_arr A B)
 | PTLamP : forall (G:ctx) (b A B:exp),
       pretyping ((th_prog,A) :: G) th_prog b B
    -> pretyping G th_log (exp_arr A B) exp_star
    -> pretyping G th_prog (exp_lam b) (exp_arr A B)
*)
 | PTRec : forall (G:ctx) (a A B:exp),
     pretyping ((th_prog,lift (exp_arr A B) 1 0) :: (th_prog,A) :: G)
            th_prog a (lift B 1 0) ->
     pretyping G th_log (exp_arr A B) exp_star ->
     pretyping G th_prog (exp_fun a) (exp_arr A B)
 | PTNRec : forall (G:ctx) (a B:exp),
     pretyping (   (th_log,exp_arr exp_nat 
                     (exp_arr (exp_eq (exp_succ (exp_var 0))
                                     (exp_var 1))
                              (lift (lift B 1 1) 1 0)))
             :: (th_log,exp_nat) :: G)
            th_log a (lift B 1 0) ->
     pretyping G th_log (exp_arr exp_nat B) exp_star ->
     pretyping G th_log (exp_nrec a) (exp_arr exp_nat B)
 | PTSub : forall (G:ctx) (a A:exp),
     pretyping G th_log a A ->
     (* typing G th_log A exp_star ->  
        or
        A = exp_star *)
     pretyping G th_prog a A
 | PTFOVal : forall (G:ctx) (v A:exp),
       val v
    -> FO A
    -> pretyping G th_prog v A
    -> pretyping G th_log A exp_star
    -> pretyping G th_log v A
 | PTNat : forall G,
     pre_ctx_wf G -> pretyping G th_log exp_nat exp_star
 | PTZero : forall (G:ctx) (t:theta),
     pre_ctx_wf G ->
     pretyping G t exp_zero exp_nat
 | PTSuc : forall (G:ctx) (t:theta) (b:exp),
     pretyping G t b exp_nat ->
     pretyping G t (exp_succ b) exp_nat

 | PTSigma : forall G A B,
      pretyping G th_log A exp_star
   -> FO A
   -> pretyping ((th_log,A) :: G) th_log B exp_star
   -> pretyping G th_log (exp_sigma A B) exp_star

 | PTPair : forall G t A B a b,
      pretyping G th_log (exp_sigma A B) exp_star
   -> pretyping G t a A
   -> pretyping G th_log ({0 ~> a}B) exp_star
   -> pretyping G t b ({0~>a}B)
   -> pretyping G t (exp_pair a b) (exp_sigma A B)

 | PTPCase : forall G t t' A1 A2 B a b,
      pretyping G t a (exp_at (exp_sigma A1 A2) t') 
   -> pretyping (   (th_log, exp_eq (exp_pair (exp_var 1) (exp_var 0)) (lift a 2 0))
              :: (t', A2) 
              :: (t', A1)
              :: G)
             t b (lift B 3 0)
   -> pretyping G th_log B exp_star
   -> pretyping G t (exp_pcase a b) B

 | PTNCase : forall (G:ctx) (t:theta) (a b1 b2 A:exp),
     pretyping G t a exp_nat ->
     pretyping ((th_log,exp_eq exp_zero a) :: G) t b1 (lift A 1 0) ->
     pretyping (   (th_log,exp_eq (exp_succ (exp_var 0)) (lift a 1 0)) 
             :: (t,exp_nat) :: G)
            t b2 (lift A 2 0) ->
     pretyping G th_log A exp_star ->
     pretyping G t (exp_ncase a b1 b2) A 
 | PTEq  : forall G a b A t1 B t2,
       pretyping G t1 a A -> pretyping G th_log A exp_star
    -> pretyping G t2 b B -> pretyping G th_log B exp_star 
    -> pretyping G th_log (exp_eq a b) exp_star
 | PTJoin : forall G a b c,
       mpstep a c -> mpstep b c
    -> pretyping G th_log (exp_eq a b) exp_star
    (* -> pretyping G t1 a A -> pretyping G t2 b B *)
    -> pretyping G th_log exp_join (exp_eq a b)
 | PTConv : forall G p a b1 b2 (* B1 B2 *) t A,
       pretyping G th_log p (exp_eq b1 b2)
(*
    -> pretyping G th_prog b1 B1 -> pretyping G th_log B1 exp_star
    -> pretyping G th_prog b2 B2 -> pretyping G th_log B2 exp_star *)
    -> pretyping G t a ({0 ~> b1}A)
    -> pretyping G th_log ({0 ~> b2}A) exp_star
    -> pretyping G t a ({0 ~> b2}A)

 | PTSubtype : forall G th a A A',
       pretyping G th a A
    -> better_typ A A'
    -> pretyping G th_log A' exp_star
    -> pretyping G th a A'
 | PTSubtype_L : forall G a A A',
       pretyping G th_log a A
    -> better_typ_L A A'
    -> pretyping G th_log A' exp_star
    -> pretyping G th_log a A'
 | PTAt  : forall G A t t',
       pretyping G th_log A exp_star
    -> pretyping G t (exp_at A t') exp_star
 | PTAtP : forall (G:ctx) (t:theta) (a A:exp),
       pretyping G t a A
    -> pretyping G th_log A exp_star
    -> pretyping G th_prog a (exp_at A t)
 | PTAtL : forall (G:ctx) (t:theta) (a A:exp),
       pretyping G th_log a A
    -> pretyping G th_log A exp_star
    -> pretyping G th_log a (exp_at A t)
 | PTAtLV : forall (G:ctx) (v A:exp),
       val v
    -> pretyping G th_prog v A
    -> pretyping G th_log A exp_star
    -> pretyping G th_log v (exp_at A th_prog).

Scheme pretyping_ind_mut := Induction for pretyping Sort Prop
  with pre_ctx_wf_ind_mut := Induction for pre_ctx_wf Sort Prop.

Hint Resolve PTVar PTApp PTLam PTRec PTZero PTSuc PTAt PTAtL
             PTAtLV PTAtP 
             PTArr PTEq PTNat PTPair PTSigma PTSub
             PTSubtype PTSubtype_L.
 
Scheme pretyping_mut_ind := Induction for pretyping Sort Prop
  with pre_ctx_wf_mut_ind := Induction for pre_ctx_wf Sort Prop.

Combined Scheme pretyping_combined_ind 
           from pretyping_mut_ind, pre_ctx_wf_mut_ind.

Hint Constructors typing.

Theorem pretyping_2_typing :
  forall G t a A,
    pretyping G t a A ->
      (typing G t a A
       /\ (typing G th_log A exp_star \/ A = exp_star)
       /\ (forall a1 a2, 
             a = (exp_eq a1 a2) 
             -> exists t1, exists A1, exists t2, exists A2,
                     typing G t1 a1 A1 /\ typing G th_log A1 exp_star
                  /\ typing G t2 a2 A2 /\ typing G th_log A2 exp_star)).  
Proof.
  apply (pretyping_ind_mut
          (fun G t a A _ =>
               (typing G t a A)
            /\ (typing G th_log A exp_star \/ A = exp_star)
            /\ (forall a1 a2, 
                  a = (exp_eq a1 a2) 
                  -> exists t1, exists A1, exists t2, exists A2,
                          typing G t1 a1 A1 /\ typing G th_log A1 exp_star
                       /\ typing G t2 a2 A2 /\ typing G th_log A2 exp_star))
          (fun G _ =>  ctx_wf G));

intros;
  repeat match goal with 
   | [ H : _ /\ _ |- _ ] => destruct H
  end;
try (split; [ idtac | split; [idtac | intros; try congruence]]);
try match goal with
  [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; destruct H
    end;
auto.

(* PTUnboxVal 
   This uses the regularity IH. *)
+ destruct H0; try congruence.
  apply typ_at_regularity in H0.
  eapply TUnboxVal; eauto.

(* PTUnboxVal regularity. *)
+ destruct H0; try congruence.
  apply typ_at_regularity in H0.
  auto.

(* PTUnboxL 
   This uses the regularity IH. *)
+ destruct H0; try congruence.
  apply typ_at_regularity in H0.
  auto.  

(* Regularity for PTUnboxL *)
+ destruct H0; try congruence.
  apply typ_at_regularity in H0.
  auto.

(* PTUnboxP 
   This uses the regularity IH. *)
+ destruct H0; try congruence.  
  apply typ_at_regularity in H0. 
  eapply TUnboxP; eauto.

(* Regularity for PTUnboxP. *)
+ destruct H0; try congruence.  
  apply typ_at_regularity in H0. 
  auto.

(* PTApp *)
+  eapply TApp; eauto.

(* PTLam *)
+ destruct t; auto.

(* PTSub *)
(* This case uses the regularity IH. *)
(* + destruct_things; subst; auto. *)


(* PTPCase *)
+ eapply TPCase; eauto.

(* PTEq *)
+ eapply TEq; eauto.

(* the eq-regularity case *)
+ (* let's just prove it manually. *)
  injection H11; intros; subst.
  exists t1.  exists A. exists t2. exists B. eauto.

(* PTJoin *)
(* This case uses the eq-regularity IH. *)
+ destruct (H1 a b eq_refl) as [t1 [A1 [t2 [ A2 [? [? [? ?]]]]]]].
  apply TJoin with (c:=c) (t1:=t1) (t2:=t2) (A:=A1) (B:=A2); auto.

(* PTConv *)
(* This case uses the regularity IH. *)
+ destruct H6; try congruence.
  apply typ_eq_regularity in H6.
  destruct H6 as [t1 [t2 [B1 [B2 [? [? [? ?]]]]]]].
  apply TConv with (p:=p) (b1:=b1) (B1:=B1) (B2:=B2).
    auto.
    apply TSubProg with (th := t1); auto.
    auto.
    apply TSubProg with (th := t2); auto.
    auto.
    auto.
    auto.

(* PTSubtype *)
+ eapply TSubtype; eauto.

(* PTSubtype_L *)
+ eapply TSubtype_L; eauto.
Qed.

Hint Constructors pretyping.
Hint Constructors pre_ctx_wf.

Theorem typing_2_pretyping :
  forall G t a A,
    typing G t a A -> pretyping G t a A.
Proof.
  apply (typing_ind_mut
          (fun G t a A _ => pretyping G t a A)
          (fun G _ =>  pre_ctx_wf G));
  intros;
  auto.

(* TUnboxVal *)
+ eapply PTUnboxVal; eauto.

(* TUnboxP *)
+ eapply PTUnboxP; eauto.

(* TApp *)
+ eapply PTApp; eauto.

(* TPCase *)
+ eapply PTPCase; eauto.

(* TEq *)
+ eapply PTEq; eauto.

(* TJoin *)
+ apply PTJoin with (c:=c); auto.
    eapply PTEq; eauto.

(* TConv *)
+ eapply PTConv; eauto.

(* TSubtype *)
+ eapply PTSubtype; eauto.

(* TSubtype_L *)
+ eapply PTSubtype_L; eauto.
Qed.

Corollary typing_iff_pretyping :
  forall G t a A,
     typing G t a A <-> pretyping G t a A.
Proof.
  split.
    apply typing_2_pretyping.
 
    intros H. apply pretyping_2_typing in H. tauto.
Qed.