Require Import List.
Require Import Arith.
Require Import Omega.
Require Import NPeano.

Require Import shared_tactics.
Require Import shared_defs.
Require Import step_defs.

(* The proofs in this file are copy-pasted from Chris' development,
   with minor alterations to make them work with changes in the definitions. *)

(* From step_db_basics.v: *)
 
Ltac extract_open_lift :=
  repeat (match goal with
  | [ |- context[ exp_fun (lift ?a 1 2) ] ] =>
    replace (exp_fun (lift a 1 2)) with (lift (exp_fun a) 1 0) in *
         by (simpl; auto)
  | [ |- context[ exp_nrec (lift ?a ?n 2) ] ] =>
    replace (exp_nrec (lift a n 2)) with (lift (exp_nrec a) n 0) in *
         by (simpl; auto)
  | [ |- context[ exp_app (lift (exp_nrec ?a) 1 0) (exp_var 1) ] ] =>
    replace (exp_app (lift (exp_nrec a) 1 0) (exp_var 1))
       with (lift (exp_app (exp_nrec a) (exp_var 0)) 1 0) in *
         by (simpl; auto)
  | [ |- context[ exp_lam (lift ?a 1 1) ] ] =>
    replace (exp_lam (lift a 1 1)) with (lift (exp_lam a) 1 0) in *
         by (simpl; auto)
  | [ |- context[ exp_arr ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] ] =>
    replace (exp_arr ({n ~> b}A) ({S n ~> lift b 1 0}B))
       with ({n ~> b}(exp_arr A B)) by (simpl; auto)
  | [ H : context[ exp_arr ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] |- _ ] =>
    replace (exp_arr ({n ~> b}A) ({S n ~> lift b 1 0}B))
       with ({n ~> b}(exp_arr A B)) in H by (simpl; auto)
  | [ H : context[ exp_arr (lift ?A 1 0) (lift ?B 1 1) ] |- _ ] =>
    replace (exp_arr (lift A 1 0) (lift B 1 1))
       with (lift (exp_arr A B) 1 0) in H by (simpl; auto)
  | [ |- context[ exp_arr (lift ?A ?n ?m) (lift ?B ?n (S ?m))] ] =>
    replace (exp_arr (lift A n m) (lift B n (S m)))
       with (lift (exp_arr A B) n m) by (simpl; auto)
  | [ H : context[ exp_arr (lift ?A ?n ?m)  (lift ?B ?n (S ?m))] |- _ ] =>
    replace (exp_arr (lift A n m) (lift B n (S m)))
       with (lift (exp_arr A B) n m) in H by (simpl; auto)
  | [ |- context[ exp_sigma ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] ] =>
     replace (exp_sigma ({n ~> b}A) ({S n ~> lift b 1 0}B))
        with ({n ~> b}(exp_sigma A B)) by (simpl; auto)
  | [ H : context[ exp_sigma ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] |- _ ] =>
     replace (exp_sigma ({n ~> b}A) ({S n ~> lift b 1 0}B))
        with ({n ~> b}(exp_sigma A B)) in H by (simpl; auto)
  | [ |- context [exp_sigma (lift ?A1 ?n ?m) (lift ?A2 ?n (S ?m))] ] =>
      replace (exp_sigma (lift A1 n m) (lift A2 n (S m))) with
              (lift (exp_sigma A1 A2) n m) by auto
  | [ H : context [exp_sigma (lift ?A1 ?n ?m) (lift ?A2 ?n (S ?m))] |- _ ] =>
      replace (exp_sigma (lift A1 n m) (lift A2 n (S m))) with
              (lift (exp_sigma A1 A2) n m) in H by auto
  | [ |- context [ exp_eq ({?n ~> ?b}?a1) ({?n ~> ?b}?a2) ] ] =>
    replace (exp_eq ({n ~> b}a1) ({n ~> b}a2))
       with ({n ~> b}(exp_eq a1 a2)) by (simpl; auto)
  | [ H : context [ exp_eq ({?n ~> ?b}?a1) ({?n ~> ?b}?a2) ] |- _ ] =>
    replace (exp_eq ({n ~> b}a1) ({n ~> b}a2))
       with ({n ~> b}(exp_eq a1 a2)) in H by (simpl; auto)
  end).

Theorem plus_comm' : forall n m, m + n = n + m.
Proof. intros; rewrite plus_comm; auto. Qed.

Theorem plus_n_Sm' : forall n m, n + S m = (S n) + m.
Proof. intros; omega.  Qed.
Ltac swap_n_Sm := rewrite plus_n_Sm'.

Theorem pred_plus_n_Sm : forall n m, pred (n + S m) = n + m.
Proof.  intros; omega.  Qed.
Hint Rewrite pred_plus_n_Sm : step.


Hint Rewrite (plus_comm' 1) : step.
Hint Rewrite (plus_comm' 2) : step.
Hint Rewrite (plus_comm' 3) : step.

Ltac destruct_typ_eq :=
  match goal with
  | [ H : {_ ~> _}?A = exp_nat |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_arr _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H :  {_ ~> _}?A = exp_at _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_eq _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_sigma _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  end.

Hint Rewrite <- plus_n_O : step.
Hint Rewrite app_length : step.


Ltac clarify_max :=
  repeat (match goal with
  | [ H : max _ _ <= _ |- _ ] =>
    apply Nat.max_lub_iff in H; destruct H
  | [ H : max _ _ < _ |- _ ] =>
    apply Nat.max_lub_lt_iff in H; destruct H
  end).

Ltac val_inversion :=
  repeat (match goal with
  | [ HV : val (exp_succ _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_ncase _ _ _) |- _ ] => inversion HV
  | [ HV : val (exp_pcase _ _) |- _ ] => inversion HV
  | [ HV : val (exp_pair _ _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_app _ _) |- _ ] => inversion HV
  end).

Ltac fo_inversion :=
  repeat (match goal with
  | [ HF : FO (exp_arr _ _) |- _ ] => inversion HF
  | [ HF : FO (exp_var _) |- _ ] => inversion HF
  | [ HF : FO (exp_sigma _  _) |- _ ] => inversion HF
  end); subst.

(* Opening and lifting closed terms/types does nothing *)
Theorem open_closed_id_gen : forall a n b,
  fvs a <= n -> {n ~> b}a = a.
Proof.
  induction a; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem open_closed_id : forall a n b,
  fvs a = 0 -> {n ~> b}a = a.
Proof.
  intros; eauto using open_closed_id_gen.
Qed.
Hint Resolve open_closed_id open_closed_id_gen.
Hint Rewrite open_closed_id_gen using (simpl; eauto 2; fail) : step.

Theorem lift_closed_id_gen : forall a n m,
  fvs a <= m -> lift a n m = a.
Proof.
  induction a; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem lift_closed_id : forall a n m,
  fvs a = 0 -> lift a n m = a.
Proof.
  intros; eauto using lift_closed_id_gen.
Qed.
Hint Rewrite lift_closed_id using (eauto 3; fail) : step.


Theorem lift_0_is_id : forall (b : exp) (m : nat),
  lift b 0 m = b.
Proof.
  induction b; intros; simpl in *; rewrite_ih; auto.
  db_var_case.
Qed. 
Hint Rewrite lift_0_is_id : step.

(* Swapping and commuting lemmas *)
Theorem lift_lift_swap_gen : forall b m1 m2 n1 n2,
    m2 <= m1
  ->   lift (lift b n1 m1) n2 m2
     = lift (lift b n2 m2) n1 (m1 + n2).
Proof.
  induction b; intros; simpl; eauto; rewrite_ih; auto;
  db_var_case.
Qed.

Theorem lift_lift_swap_0 : forall b m1 n1 n2,
       lift (lift b n1 m1) n2 0
     = lift (lift b n2 0) n1 (n2 + m1).
Proof.
  intros; subst; rewrite lift_lift_swap_gen; eauto with arith.
Qed.
Hint Rewrite <- lift_lift_swap_0 : step.
Hint Extern 4 =>
  match goal with
  | [ H : context [ lift (lift ?A 1 1) ?n (S ?m) ]
     |- context [lift (lift ?A ?n ?m) 1 1] ] =>
    rewrite lift_lift_swap_gen with (b := A) by omega
  | [ |- context[ lift (exp_fun (lift _ 1 2)) _ (S ?m) ] ] =>
    extract_open_lift; replace (S m) with (1 + m) by auto;
    rewrite <- lift_lift_swap_0
  | [ |- context[ lift (exp_nrec (lift _ 1 2)) _ (S ?m) ] ] =>
    extract_open_lift; replace (S m) with (1 + m) by auto;
    rewrite <- lift_lift_swap_0
  | [ |- context[lift (lift _ _ ?m1) ?n2 0] ] =>
    rewrite lift_lift_swap_0
  | [ H : context[lift (lift _ _ ?m1) ?n2 0] |- _ ] =>
    rewrite lift_lift_swap_0 in H
  end.

Theorem lift_lift_collapse : forall b n1 n2 m1 m2,
     m1 <= m2
  -> m2 <= m1 + n1
  ->   lift (lift b n1 m1) n2 m2
     = lift b (n1 + n2) m1.
Proof.
  induction b; intros; simpl in *; auto; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_lift_collapse using omega : step.

Theorem lift_lift_collapse_1 : forall b n1,
       lift (lift b n1 0) 1 0
     = lift b (S n1) 0.
Proof.
  intros; autorewrite with step; simpl; auto.
Qed.
Hint Rewrite lift_lift_collapse_1 : step.

Theorem lift_open_commute : forall B b i n m,
     i <= m
  ->   lift ({i ~> b}B) n m
     = {i ~> lift b n m}(lift B n (S m)).
Proof.
  induction B; intros; simpl in *; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_open_commute using omega : step.
Hint Extern 4 =>
  match goal with
  | [ |- context [lift ({_ ~> _}_) _ _] ] =>
    rewrite lift_open_commute by (auto with arith)
  end.

Theorem open_lift_commute : forall b a n1 n2 i m,
     m <= n2
  -> m <= i 
  -> {n1 + i ~> (lift a (n1 + n2) 0)} (lift b n1 m) 
   = lift ({i ~> lift a n2 0} b) n1 m.
Proof.
  induction b; intros; simpl in *; eauto;
  autorewrite with step; simpl in *;
  repeat (match goal with
  | [ |- context [ S (_ + _) ] ] =>
    repeat (rewrite plus_n_Sm)
  | [ |- context [ (_ + _ + _) ] ] =>
    repeat (rewrite <- plus_assoc)
  end); rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite <- open_lift_commute using omega : step.

Theorem open_lift_commute_1 : forall b a n2 i m,
     m <= n2 -> m <= i
  ->   {S i ~> (lift a (S n2) 0)}(lift b 1 m)
     = lift ({i ~> lift a n2 0} b) 1 m.
Proof.
  intros b a n2 i m ? ?; replace (S i) with (1 + i) by auto;
  autorewrite with step in *; eauto.
Qed.
Hint Rewrite <- open_lift_commute_1 using omega : step.

Theorem open_lift_commute_0 : forall b a n1,
    {n1 ~> lift a n1 0} (lift b n1 0)
  = lift ({0 ~> a} b) n1 0.
Proof.
  intros b a n1; replace ({0 ~> a} b) with ({0 ~> lift a 0 0} b);
  try rewrite <- open_lift_commute by omega;
  autorewrite with step; eauto.
Qed.

Theorem open_lift_commute_1_closed : forall b a i m,
     m <= i
  -> fvs a = 0
  ->   {S i ~> a} (lift b 1 m) 
     = lift ({i ~> a} b) 1 m.
Proof.
  intros b a i m H1 HFV;
  replace ({i ~> a}) with ({i ~> lift a m 0})
       by (autorewrite with step; eauto);
  replace ({S i ~> a}) with ({S i ~> lift a (S m) 0})
       by (autorewrite with step; eauto);
  autorewrite with step; eauto.
Qed.
Hint Rewrite <- open_lift_commute_1_closed using (simpl; omega): step.

Theorem lift_ctx_length : forall n m G,
  length G = length (lift_c G n m).
Proof.
  induction G as [ | [? ?] ]; intros; simpl in *; eauto.
Qed.
Hint Rewrite <- lift_ctx_length : step.
Hint Resolve lift_ctx_length.

Theorem lift_ctx_app : forall G1 G2 n m,
    lift_c (G1 ++ G2) n m 
  = (lift_c G1 n (m + length G2)) ++ lift_c G2 n m.
Proof.
  induction G1 as [ | [? ?] ? ]; intros; simpl in *; eauto;
  try rewrite app_length; repeat (f_equal; try omega; eauto).
Qed.
Hint Rewrite lift_ctx_app : step.


Theorem open_under_lift : forall b a i n m,
     i <= n + m
  -> m <= i
  -> {i ~> a} (lift b (S n) m) = lift b n m.
Proof.
  induction b; intros; simpl in *; rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite open_under_lift using omega : step.

Theorem open_into_lift : forall A n,
  {n ~> exp_var (S n)} (lift A 2 (S n)) = lift A 1 n.
Proof.
  induction A; intros; simpl; eauto;
  try solve [autorewrite with step; repeat rewrite_ih; eauto];
  db_var_case.
Qed.

Theorem open_open_swap : forall b n1 n2 a1 a2,
    n2 <= n1
  ->   {n1 ~> a1}({n2 ~> a2}b)
     = {n2 ~> ({n1 ~> a1}a2)}({S n1 ~> lift a1 1 n2}b).
Proof.
  induction b; intros; simpl in *; rewrite_ih; eauto;
  try match goal with
  | [ |- _ ] =>
    (* a little bit of fanciness so that we can see open_lift_commute applies *)
    f_equal; auto; match goal with
    | [ |- context[lift ({?i ~> ?a}?b) ?n 0] ] =>
      replace (lift ({i ~> a}b) n 0) with (lift ({i ~> lift a 0 0}b) n 0)
        by (rewrite lift_0_is_id; eauto)
    end; rewrite <- open_lift_commute; eauto
  end;
  try db_var_case.
Qed.


(* The de bruijn operations meet FO *)
Theorem lift_preserves_fo : forall A,
  FO A -> forall n m, FO (lift A n m).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve lift_preserves_fo.

Theorem open_preserves_fo : forall A,
  FO A -> forall a n, FO ({n ~> a} A).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve open_preserves_fo.

Theorem lifted_fo : forall A n m,
  FO (lift A n m) -> FO A.
Proof.
  induction A; inversion 1; subst; eauto.
Qed.
Hint Extern 1 =>
  match goal with
  | [ H : FO (lift ?A ?n ?m) |- FO ?A ] =>
    eapply lifted_fo; eexact H
  end.

Theorem pstep_preserves_fo : forall A,
  FO A -> forall B, pstep A B -> FO B.
Proof.
  induction 1; eauto; inversion 1; subst; eauto.
Qed.
Hint Immediate pstep_preserves_fo.

(* the de bruijn operations meet head_form *)
Theorem lift_preserves_hf : forall A h,
  head_form A h -> forall n m, head_form (lift A n m) h.
Proof.
  inversion 1; intros; simpl in *; subst; eauto.
Qed.
Hint Resolve lift_preserves_hf.

Theorem open_preserves_hf : forall A h,
  head_form A h -> forall n u, head_form ({n ~> u}A) h.
Proof.
  inversion 1; intros; simpl in *; subst; eauto.
Qed.
Hint Resolve open_preserves_hf.

(* The de bruijn operations meet reduction *)
Theorem lift_preserves_val : forall v,
  val v -> forall n m, val (lift v n m).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve lift_preserves_val.

Theorem open_preserves_val : forall v,
  val v -> forall v' n, val v' -> val ({n ~> v'} v).
Proof.
  induction 1; intros; simpl in *; eauto;
  clarify_compares; simpl; eauto.
Qed.
Hint Resolve open_preserves_val.

Theorem pstep_preserves_val : forall v v',
  pstep v v' -> val v -> val v'.
Proof.
  induction 1; intros; subst; val_inversion; eauto.
Qed.
Hint Immediate pstep_preserves_val.

Theorem lift_preserves_pstep : forall a b,
     pstep a b
  -> forall n m, 
       pstep (lift a n m) (lift b n m).
Proof.
  induction 1; intros; simpl in *; eauto;
  match goal with
  (* the natind case *)
  | [ |- pstep _ (lift ({0 ~> _} ({0 ~> exp_lam (exp_lam _)} _)) _ ?m) ] =>
    repeat (rewrite lift_open_commute by auto); simpl;
    replace (S (S (S (S (S m))))) with (S (S m) + 3) by omega;
    rewrite <- lift_lift_swap_gen by omega; eauto

  | [ |- pstep _ (lift ({0 ~> _} ({0 ~> _} ?a)) _ ?m) ] 
    => repeat (rewrite lift_open_commute by auto); eauto;
       extract_open_lift; replace (S m) with (1 + m) by (simpl; auto);
       rewrite <- lift_lift_swap_0; simpl; auto
                       
  end.
Qed.
Hint Resolve lift_preserves_pstep.

Theorem lift_preserves_mpstep : forall a b,
    mpstep a b
 -> forall n m,
      mpstep (lift a n m) (lift b n m).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve lift_preserves_mpstep.

Theorem open_preserves_pstep_r :  forall a a',
  pstep a a' -> forall v n, val v -> pstep ({n ~> v}a) ({n ~> v}a').
Proof.
  induction 1; intros; simpl in *; eauto 7;
  rewrite open_open_swap with (n2 := 0) by omega; eauto;
  try (rewrite open_open_swap with (n2 := 0) by omega; eauto);
  simpl; autorewrite with step; simpl;
  try rewrite open_lift_commute_1 by omega; eauto 6.

(* The PCasePair case, sorry for not automating more. *)
rewrite open_open_swap with  (a2 := exp_join) (n2:=0) by omega;
rewrite lift_0_is_id; 
rewrite open_closed_id with (a := exp_join) by auto;
rewrite lift_lift_collapse by auto; auto.

(* natrec *)
  replace (S (S (S (S (S n))))) with (3 + (S (S n))) by omega.
  replace 5 with (3 + 2) by omega.
  rewrite open_lift_commute by omega.  eauto 6.
Qed.
Hint Resolve open_preserves_pstep_r.

Theorem open_preserves_mpstep_r : forall a a',
  mpstep a a' -> forall v n, val v -> mpstep ({n ~> v}a) ({n ~> v}a').
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve open_preserves_mpstep_r.

Theorem open_preserves_pstep_l : forall b a a',
  pstep a a' -> forall n, pstep ({n ~> a}b) ({n ~> a'}b).
Proof.
  induction b; intros; simpl in *; eauto 7;
  db_var_case.
Qed.
Hint Resolve open_preserves_pstep_l.

(* Variable numbers don't really matter *)
Theorem open_index_swap : forall a n1 n2,
  exists a', forall u, {n1 ~> u}a = {n2 ~> u}a'.
Proof.
  induction a; intros; simpl in *; eauto;
  try match goal with
  | [ |- context[nat_compare _ _] ] => 
    clarify_compares; subst; eauto;
    match goal with
    | [ |- exists _, forall u, u = {?n2 ~> u} _ ] =>
      exists (exp_var n2); db_var_case
    | [ |- exists _, forall _, exp_var ?n = {?n2 ~> _} _ ] =>
      destruct (le_gt_dec n2 n); 
        [ exists (exp_var (S n)); simpl; destruct n2; db_var_case
        | exists (exp_var n); simpl; db_var_case ]
    end
  | [ IHa : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _,
      IHb : forall _ _, exists _, forall _, ({_ ~> _} ?b) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) ({_ ~> _} ?b) = _]  =>
    let a' := fresh a in let b' := fresh b in
    edestruct IHa as [a' ?]; edestruct IHb as [b' ?];
    exists (P a' b'); simpl; intros; f_equal; eauto
  | [ IHa : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _,
      IHb : forall _ _, exists _, forall _, ({_ ~> _} ?b) = _,
      IHc : forall _ _, exists _, forall _, ({_ ~> _} ?c) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) ({_ ~> _} ?b) ({_ ~> _} ?c) = _] =>
    let a' := fresh a in let b' := fresh b in let c' := fresh c in
    edestruct IHa as [a' ?]; edestruct IHb as [b' ?]; edestruct IHc as [c' ?];
    exists (P a' b' c'); simpl; intros; f_equal; eauto
  | [ IH : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) = _]  =>
    let a' := fresh a in 
    edestruct IH as [a' ?];
    exists (P a'); simpl; intros; f_equal; eauto
  | [ IH : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) ?t = _]  =>
    let a' := fresh a in 
    edestruct IH as [a' ?];
    exists (P a' t); simpl; intros; f_equal; eauto
  | [ |- exists _, forall _, ?a = _ ] => 
    exists a; intros; eauto
  end.
Qed.

(* Lemmas about open_ctx *)
Theorem open_c_length : forall n v G,
  length (open_c n v G) = length G.
Proof.
  induction G as [ | [ ? ? ] ? ]; simpl in *; eauto.
Qed.
Hint Rewrite open_c_length : step.

Theorem open_c_app : forall n v G1 G2,
    open_c n v (G1 ++ G2) 
  = (open_c (n + length G2) v G1) ++ (open_c n v G2).
Proof.
  induction G1 as [ | [ ? ? ] ? ]; intros; simpl in *; eauto; 
  rewrite_ih; autorewrite with step in *; 
  repeat (f_equal; eauto; try omega).
Qed.
Hint Rewrite open_c_app : step.

Theorem open_open_swap_c : forall G a1 a2,
     fvs a1 = 0
  ->   open_c 0 a1 (open_c 0 a2 G)
     = open_c 0 ({0 ~> a1}a2) (open_c 1 a1 G).
Proof.
  induction G as [ | [? ?] ? ]; intros; simpl in *; auto;
  try solve
    [rewrite_ih; try rewrite <- open_lift_commute_0;
     rewrite open_open_swap by (autorewrite with step; eauto);
     autorewrite with step; eauto].
Qed.



(* The following is copy-pasted from Chris' step_typing_basics.v: *)

Scheme typing_mut_ind := Induction for typing Sort Prop
  with ctx_wf_mut_ind := Induction for ctx_wf Sort Prop.

Combined Scheme typing_combined_ind 
           from typing_mut_ind, ctx_wf_mut_ind.

Hint Rewrite <- app_assoc : step.

Theorem TNat_gen : forall G t,
  ctx_wf G -> typing G t exp_nat (exp_star).
Proof.
  destruct t; intros; eauto using TSub, TSubType.
Qed.
Hint Extern 1 =>
  match goal with
  | [ |- typing ?G _ exp_nat _ ] =>
    first [ apply TNat | apply TNat_gen ]
  end.

Ltac inst_equalities := 
  repeat (match goal with
  | [ H : forall _ _, _ ++ _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ ++ _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, (_ ++ _ :: _) ++ _ :: _ = _ -> _ |- _ ] =>
    try rewrite <- app_assoc in H; try rewrite <- app_comm_cons in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H

  | [ H : forall _ _ _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall G, ctx_wf (G ++ ?G3) -> _,
      H2 : ctx_wf (?G2 ++ ?G3)
     |- context[lift_c _ (length ?G2) _] ] => let Hwf := fresh "Hwf" in 
    assert (Hwf := H _ H2); clear H; rename Hwf into H 
  | [ H : forall _ _ _ _, _ ++ _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, (_ ++ _ :: _) ++ _ :: _ = _ -> _ |- _ ] =>
    try rewrite <- app_assoc in H; try rewrite <- app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H

  | [ H : forall _ _ _ _ _ _ _, _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ ++ _ :: ?G = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing ?G _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H

  | [ H : forall _ _ _ _ _ _ _, _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ :: _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ ++ _ :: _ ++ (_,({_ ~> _}_)) :: _
                              = _ ++ (_,({_ ~> _}_)) :: _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, (_ ++ (_,({_ ~> _}_)) :: _) ++ _
                              = _ ++ (_,({_ ~> _}_)) :: _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    rewrite <- app_assoc in H; rewrite <- app_comm_cons in H;
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ :: _ :: _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  end).

(* These two theorems invert some equalities on lists, given the types of
   information we get out of the clarify_compares operation *)
Theorem app_eq_lengths_eq : forall {A} (G1 G2 G3 G4 : list A),
     G1 ++ G2 = G3 ++ G4
  -> length G1 = length G3
  -> G1 = G3 /\ G2 = G4.
Proof.
  induction G1; intros G2 G3 G4 Heq Hlen;
  destruct G3; simpl in *; inversion Hlen; auto;
  inversion Heq; subst; eauto.
  match goal with
  | [ IH : forall _ _ _, ?G1 ++ _ = _ -> length ?G1 = _ -> _ /\ _,
      Heq : ?G1 ++ _ = _,
      Hlen : length ?G1 = _ |- _ ] =>
      destruct (IH _ _ _ Heq Hlen); subst; eauto
  end.
Qed.

Theorem app_eq_lengths_lt : forall {A} (G1 G2 G3 G4 : list A),
     G1 ++ G2 = G3 ++ G4
  -> length G1 < length G3
  -> exists G5, G3 = G1 ++ G5 /\ G2 = G5 ++ G4.
Proof.
  induction G1; intros G2 G3 G4 Heq Hlen;
  destruct G3; simpl in *; subst; eauto;
  try solve [inversion Hlen];
  inversion Heq; subst;
  match goal with
  | [ IH : forall _ _ _, _ ++ _ = _ ++ _ -> _ -> exists _ , _,
      Heq : _ ++ _ = _ ++ _ |- _ ] =>
    destruct (IH _ _ _ Heq) as [G5 [Heq1 Heq2]]
  end; try omega; subst; eauto.
Qed.

(* This tactic just tries to apply the previous two theorems. *)
Ltac invert_list_equalities :=
  (* Begin by making the > case look like the < case. *)
  try match goal with
  | [ H1 : length ?G2 > length ?G1, H2 : ?G2 ++ _ = ?G1 ++ _ |- _ ] =>
    unfold gt in H1; apply eq_sym in H2
  end;
  match goal with
  | [ Heq : ?G1 ++ _ = ?G2 ++ _, Hlen : length ?G1 = length ?G2 |- _ ] =>
    destruct (app_eq_lengths_eq _ _ _ _ Heq Hlen); subst; clear Hlen Heq
  | [ Heq : ?G1 ++ _ = ?G2 ++ _, Hlen : length ?G1 < length ?G2 |- _ ] =>
    let G5 := fresh "G5" in
    let HEq1 := fresh "HEq1" in let HEq2 := fresh "HEq2" in
    destruct (app_eq_lengths_lt _ _ _ _ Heq Hlen) as [G5 [Heq1 Heq2]]; subst
  end.

(* this one looks for the situation where we know:

     length G1 < length (G1 ++ G5)

   Then G5 must be a cons, so we expose it *)
Ltac find_cons := 
  match goal with
  | [ Hlen : length ?G1 < length (?G1 ++ ?G5) |- _ ] =>
    destruct G5; 
      [  subst; rewrite app_length in Hlen; rewrite plus_comm in Hlen; 
         simpl in Hlen; destruct (lt_irrefl _ Hlen) 
       | subst; clear Hlen ];
    match goal with
    | [ Heq2 : _ :: _ = (_ :: _) ++ _ |- _ ] =>
      inversion Heq2; subst; clear Heq2
    end
  | [ HConsEq : _ :: _ = _ :: _ |- _ ] =>
    inversion HConsEq; subst; clear HConsEq
  end.

Theorem ctx_wf_inversion_app : forall G1 G2,
  ctx_wf (G1 ++ G2) -> ctx_wf G2.
Proof.
  induction G1; intros; simpl in *; auto; match goal with
  | [ H : ctx_wf _ |- _ ] => inversion H
  end; subst; auto.
Qed.
Hint Immediate ctx_wf_inversion_app.

Ltac notProvenYet P :=
  match goal with
  | [ _ : P |- _ ] => fail 1
  | _ => idtac
  end.

Ltac invert_ctx_wf := simpl in *;
  try match goal with
  | [ H : ctx_wf (_ ++ ?G) |- _ ] =>
    notProvenYet (ctx_wf G); let HWF := fresh "HWF" in
    assert (HWF := ctx_wf_inversion_app _ _ H); invert_ctx_wf
  | [ H : ctx_wf ((_,?A) :: ?G) |- _ ] =>
     notProvenYet(ctx_wf G); inversion H; subst; invert_ctx_wf
  | _ => idtac
  end.

Theorem typing_regularity_ctx : forall G t a A,
  typing G t a A -> ctx_wf G.
Proof.
  induction 1;  invert_ctx_wf; eauto.
Qed.
Hint Immediate typing_regularity_ctx.



(**** Free variables ***)
Hint Resolve Nat.max_lub.
Hint Resolve Nat.max_lub_l.
Hint Resolve Nat.max_lub_r.

Theorem typing_fvs : 
  forall G t a A, typing G t a A
    -> fvs a <= length G.
Proof.
  induction 1; intros; auto; subst; simpl in *;
  autorewrite with step in *; simpl in *; eauto; omega.
Qed.

Theorem typing_fvs_closed_exp : forall t a A,
  typing nil t a A -> fvs a = 0.
Proof.  
  intros t a A H; apply typing_fvs in H; simpl in *; omega.
Qed.
Hint Immediate typing_fvs_closed_exp.


(*** Derivable generalized versions of some typing rules ****)
Lemma TLamGen : forall (G:ctx) th (b A B:exp),
                  typing ((th,A) :: G) th b B
    -> typing G th_log (exp_arr A B) exp_star
    -> typing G th (exp_lam b) (exp_arr A B).
Proof.
  intros. destruct th; [apply TLam | apply TLamP]; auto.
Qed.

Lemma TFOVal' : forall (G:ctx) (v A:exp) th th',
       val v
    -> FO A
    -> typing G th v A
    -> typing G th_log A exp_star
    -> typing G th' v A.
Proof.
  intros.
  destruct th; destruct th'; auto.
Qed.
Hint Resolve TFOVal'.  

(*** lift meats the subtyping relation. ***)

Lemma lift_better_typ : forall A A',
  better_typ A A'
 -> forall n m, better_typ (lift A n m) (lift A' n m).
Proof.
Hint Constructors better_typ.
  induction 1; simpl; auto.
Qed.
Hint Resolve lift_better_typ.

Lemma lift_better_typ_L : forall A A',
  better_typ_L A A'
 -> forall n m, better_typ_L (lift A n m) (lift A' n m).
Proof.
Hint Constructors better_typ_L.
  induction 1; simpl; auto.
Qed.
Hint Resolve lift_better_typ_L.

Lemma not_better_star2 : forall A A',
     better_typ A A'
  -> A = exp_star
  -> False.
Proof.
  induction 1; 
  try match goal with 
    [ H : FO _ |- _ ] => inversion H; intros
  end;
  congruence.
Qed.

Lemma not_better_L_star2 : forall A A',
     better_typ_L A A'
  -> A = exp_star
  -> False.
Proof.
  induction 1; try congruence; eauto using not_better_star2.
Qed.


Theorem kinding_subsumption_gen : forall G t a,
  typing G th_log a exp_star -> typing G t a exp_star.
Proof.  intros; destruct t; eauto using TSubType.  
Qed.
Hint Immediate kinding_subsumption_gen.


(****************************************************************************)
(* from step_preservation.v *)


Theorem regularity : forall G t a A,
     typing G t a A
  ->    (typing G th_log A exp_star)
     \/ (A = exp_star).
Proof.
  induction 1; invert_ctx_wf; destruct_things; eauto.
Qed.


Theorem typ_eq_regularity_gen : forall G t a A,
     typing G t a A 
  -> forall b1 b2, a = exp_eq b1 b2
      -> exists t1, exists t2, exists B1, exists B2,
             typing G t1 b1 B1 /\ typing G th_log B1 exp_star
          /\ typing G t2 b2 B2 /\ typing G th_log B2 exp_star.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_eq _ _ |- _] => inversion H; subst; clear H
  end; eauto; try destruct_ih; eauto 10.
Qed.

Theorem typ_eq_regularity : forall G t b1 b2 A,
     typing G t (exp_eq b1 b2) A
      -> exists t1, exists t2, exists B1, exists B2,
             typing G t1 b1 B1 /\ typing G th_log B1 exp_star
          /\ typing G t2 b2 B2 /\ typing G th_log B2 exp_star.
Proof.
  intros; eauto using typ_eq_regularity_gen.
Qed.

Theorem typ_at_regularity_gen : forall G t a A,
     typing G t a A
  -> forall t' A', a = exp_at A' t'
      -> typing G th_log A' exp_star.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_at _ _ |- _] => inversion H; subst; clear H
  end; eauto.
Qed.

Theorem typ_at_regularity : forall G t B t' A,
     typing G t (exp_at B t') A
  -> typing G th_log B exp_star.
Proof.
  intros; eauto using typ_at_regularity_gen.
Qed.


(*** Some more derivable generalized typing rules, these ones use regularity. ***)

Lemma TSubProg : forall G th a A,
  typing G th a A -> typing G th_prog a A.
Proof.
  intros.
  destruct (regularity _ _ _  _ H).
    + destruct th; auto.
    + subst.  destruct th;  auto. 
Qed.
Hint Resolve TSubProg.

Lemma TSub' : forall G th a A,
  typing G th_log a A -> typing G th a A.
Proof.
  intros.
  destruct (regularity _ _ _  _ H).
    + destruct th; auto.
    + subst.  destruct th;  auto. 
Qed.
Hint Resolve TSub'.


(******************************************************************)
(*             Inversion                                          *)
(******************************************************************)

Theorem star_has_no_type' : forall G t a A,
     typing G t a A
  -> a = exp_star
  -> False.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_star |- _] => inversion H; subst; clear H
  end; eauto.
Qed.  

Theorem star_has_no_type : forall G t A,
   typing G t exp_star A -> False.
Proof.
  intros; eauto using star_has_no_type'.
Qed.

Theorem eq_star : forall b A,
     {0 ~> b}A = exp_star
  -> (A = exp_star) \/ (b = exp_star /\ A = exp_var 0).
Proof.
  destruct A; intros; subst; eauto;
  (match goal with
  | [ H : _ = _ |- _ ] =>
    simpl in H; try solve [inversion H];
    db_var_case; inversion H
  end).
Qed.


Ltac star_isnt_equal := 
  match goal with
  | [ H : typing _ th_log _ (exp_eq _ exp_star) |- _ ] =>
       apply regularity in H; destruct H as [ H | H ]; [ | inversion H];
       eapply typ_eq_regularity in H; eauto; destruct_things;
       match goal with
       | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
       | [ H : typing _ _ exp_star _ |- _ ] =>
         apply star_has_no_type in H; destruct H
       end
  | [ H : typing _ th_log _ (exp_eq exp_star _) |- _ ] =>
       apply regularity in H; destruct H as [ H | H ]; [ | inversion H];
       eapply typ_eq_regularity in H; eauto; destruct_things;
       match goal with
       | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
       | [ H : typing _ _ exp_star _ |- _ ] =>
         apply star_has_no_type in H; destruct H
       end
  | [ H : typing _ th_log _ (exp_eq _ (exp_at (exp_star) _)) |- _ ] =>
    apply regularity in H; (destruct H as [ H | H ]; [ | inversion H]);
    eapply typ_eq_regularity in H; eauto; destruct_things;
    match goal with
    | [ H : typing _ _ (exp_at exp_star _) _ |- _ ] =>
      apply typ_at_regularity in H; destruct_things;
      apply star_has_no_type in H; destruct H
    end
  | [ H : typing _ _ exp_star _ |- _ ] =>
      apply star_has_no_type in H; destruct H

  | [ H1 : typing _ _ ?b _,
      H2 : ?b = exp_star |- _ ] =>
      rewrite H2 in H1; apply star_has_no_type in H1; destruct H1

  end.

Theorem typ_sigma_inversion_gen : forall G t a A,
     typing G t a A
  -> forall B1 B2, a = exp_sigma B1 B2
     -> (   FO B1
         /\ A = exp_star
         /\ typing G th_log B1 exp_star
         /\ typing ((th_log,B1)::G) th_log B2 exp_star).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_sigma _ _  |- _] => inversion H; subst; clear H
  end; val_inversion; first [destruct_ih; subst | eauto 10]; eauto;
  try match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; inversion H
  | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
  | [ H : exp_at _ _ = exp_star |- _ ] => inversion H
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; eauto; star_isnt_equal
  | [ H : better_typ exp_star _ |- _ ] =>
    elimtype False; eauto using not_better_star2     
  | [ H : better_typ_L exp_star _ |- _ ] =>
    elimtype False; eauto using not_better_L_star2     
  end.
Qed.

Theorem typ_sigma_inversion : forall G t A1 A2 B,
     typing G t (exp_sigma A1 A2) B
  ->     FO A1
      /\ B = exp_star
      /\ typing G th_log A1 exp_star
      /\ typing ((th_log,A1)::G) th_log A2 exp_star.
Proof.
  intros G t A1 A2 B H; eapply typ_sigma_inversion_gen in H; eauto;
  destruct_things; eauto.
Qed.


Theorem typ_eq_inversion_gen : forall G t a A,
     typing G t a A
  -> forall b1 b2, a = exp_eq b1 b2
     -> exists B1, exists t1, exists B2, exists t2,
            A = exp_star
         /\ typing G t1 b1 B1
         /\ typing G t2 b2 B2.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_eq _ _  |- _] => inversion H; subst; clear H
  end; val_inversion; first [destruct_ih; subst | eauto 10]; eauto 10;
  try match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; inversion H
  | [ H : exp_eq _ _ = exp_star |- _ ] => inversion H
  | [ H : exp_at _ _ = exp_star |- _ ] => inversion H
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; simpl in *; eauto; star_isnt_equal
  | [ H : better_typ exp_star _ |- _ ] =>
    elimtype False; eauto using not_better_star2     
  | [ H : better_typ_L exp_star _ |- _ ] =>
    elimtype False; eauto using not_better_L_star2     
  end.
Qed.

Theorem typ_eq_inversion : forall G t A b1 b2,
     typing G t (exp_eq b1 b2) A
  -> exists B1, exists t1, exists B2, exists t2, 
        A = exp_star
     /\ typing G t1 b1 B1
     /\ typing G t2 b2 B2.
Proof.
  intros G t A1 A2 B H; eapply typ_eq_inversion_gen in H; eauto;
  destruct_things; eauto.
Qed.

Theorem typ_arr_inversion_gen : forall G t a A,
     typing G t a A
  -> forall B1 B2, a = exp_arr B1 B2
     ->    FO B1
        /\ A = exp_star
        /\ typing G th_log B1 exp_star
        /\ typing ((th_log,B1)::G) th_log B2 exp_star.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_arr _ _  |- _] => inversion H; subst; clear H
  end; eauto; try (destruct_ih; subst; (split; [ auto | ]));
  try match goal with
  | [ H : ?A = ?B |- _ ] => solve [inversion H]
  | [ H : typing _ _ exp_star _ |- _ ] => apply star_has_no_type in H; inversion H
  | [ H : better_typ exp_star _ |- _ ] =>
    elimtype False; eauto using not_better_star2     
  | [ H : better_typ_L exp_star _ |- _ ] =>
    elimtype False; eauto using not_better_L_star2     
  end; eauto; 
  try match goal with
  | [ H : {_ ~> _}_ = exp_star |- _ ] =>
    apply eq_star in H; destruct_things; subst; eauto; star_isnt_equal
  end.
Qed.

Theorem typ_arr_inversion : forall G t A1 A2 B,
     typing G t (exp_arr A1 A2) B
  ->    FO A1
     /\ B = exp_star
     /\ typing G th_log A1 exp_star
     /\ typing ((th_log,A1)::G) th_log A2 exp_star.
Proof.
  intros G t A1 A2 B H; eapply typ_arr_inversion_gen in H; eauto;
  destruct_things; eauto.
Qed.

Ltac destruct_things :=
  repeat (match goal with
  | [ H : exists _, _ |- _ ] => destruct H
  | [ H : _ /\ _ |- _ ] => destruct H
  | [ H : _ \/ _ |- _ ] => destruct H
  end).

Ltac untyped a :=
  match goal with
  | [ H : typing _ _ a _ |- _ ] => fail 1
  | _ => idtac
  end.

(* A cut-down version of Chris' tactic of the same name. *)
Ltac go_go_regularity :=
  simpl in *;  
  repeat (match goal with
  | [ H : typing _ _ _ ?A |- _ ] =>
    let HNeq := fresh "HNeq" in let HContra := fresh "HContra" in
    let HReg := fresh "HReg" in
    untyped A; assert (A = exp_star -> False) as HNeq
                   by (intro HContra; inversion HContra);
    destruct (regularity _ _ _ _ H) as [HReg | HReg];
    [ | apply HNeq in HReg; destruct HReg]; clear HNeq
  end);
  repeat (match goal with
  | [ H : typing ?G _ (exp_at ?b _) _ |- _ ] =>
    untyped b; let H' := fresh H in 
    assert (H' := typ_at_regularity _ _ _ _ _ H)
  | [ H : typing ?G ?t (exp_arr ?A ?B) _ |- _ ] =>
    first [untyped A | untyped B]; let H' := fresh H in
    assert (H' := typ_arr_inversion _ _ _ _ _ H); destruct_things; subst
  | [ H : typing ?G ?t (exp_arr ?A ?B) ?x |- _ ] =>
    is_var x; let H' := fresh H in
    assert (H' := typ_arr_inversion _ _ _ _ _ H);
    destruct H' as [? [? [_ _]]]; subst
  | [ H : typing _ _ (exp_sigma ?A ?B) _ |- _ ] =>
    first [untyped A | untyped B]; let H' := fresh H in
    assert (H' := typ_sigma_inversion _ _ _ _ _ H);
    destruct_things; subst
  | [ H : typing ?G _ (exp_eq ?a ?b) _ |- _ ] =>
    let t := fresh "t" in 
    first 
      [  untyped a; untyped b; let H' := fresh H in 
         assert (H' := typ_eq_regularity _ _ _ _ _ H); destruct_things
      | untyped a; destruct (typ_eq_regularity _ _ _ _ _ H)
                         as [? [t [? [_ [? [? [? ?]]]]]]]; clear t
      | untyped b; destruct (typ_eq_regularity _ _ _ _ _ H)
                         as [t [? [_ [? [_ [? [? ?]]]]]]]; clear t
      ]
  end).


(*************************************************)  
(****************** Weakening ********************)
(* 
Theorem weakening_combined :
     (   forall G t a A, typing G t a A 
      -> forall G1 G3, G = G1 ++ G3
      -> forall G2, let n := length G2 in let m := length G1 in
              ctx_wf (G2 ++ G3)
           -> typing (lift_c G1 n 0 ++ G2 ++ G3) t
                     (lift a n m) (lift A n m))
  /\ (   forall G, ctx_wf G
      -> forall G1 G3, G = G1 ++ G3
      -> forall G2, ctx_wf (G2 ++ G3)-> 
           let n := length G2 in
           ctx_wf ((lift_c G1 n 0) ++ G2 ++ G3)).
Proof.
  apply typing_combined_ind;
  intros; subst; eauto;
  repeat (match goal with | [n := length _ : nat |- _] => subst n end);
  simpl in *;
  try solve [inst_equalities; mycrush; simpl; extract_open_lift; eauto].
(*
  try match goal with
  (* the ctx_wf cases *)
  | [ HEq : _ = ?G1 ++ _ |- ctx_wf _ ] =>
    destruct G1; inversion HEq; simpl in *; inst_equalities; eauto
  (* The var case *)
  | [ |- typing _ _ (exp_var _) _ ] =>
    clarify_compares; subst; eauto; invert_list_equalities;
    [ autorewrite with step; rewrite app_assoc;
      apply TVar; autorewrite with step; eauto
  
    | find_cons;  autorewrite with step in *; simpl; eauto;
      inst_equalities; match goal with 
      | [ |- context [ ?n + S ?m ] ] =>
        replace (n + S m) with ((S n) + m) by omega
      end; autorewrite with step in *; simpl in *; eauto
  
    | autorewrite with step; repeat (rewrite app_assoc); apply TVar;
      autorewrite with step; try omega; eauto ]
  end; 
*)

(* var *)
+  clarify_compares; subst; eauto; invert_list_equalities.
  *  autorewrite with step; rewrite app_assoc;
      apply TVar; autorewrite with step; eauto.
     assert (Hwellformed := H0 G0 ((t,A)::G3) eq_refl G4);
       autorewrite with step in Hwellformed; auto.
  *  find_cons;  autorewrite with step in *; simpl; eauto;
      inst_equalities; match goal with 
      | [ |- context [ ?n + S ?m ] ] =>
        replace (n + S m) with ((S n) + m) by omega
      end; autorewrite with step in *; simpl in *; eauto.

     (* Hm??? What is going on here?? *)
     admit.

  *  autorewrite with step; repeat (rewrite app_assoc); apply TVar;
      autorewrite with step; try omega; eauto.

     assert ((G0 ++ G5) ++ (t, A) :: G3 = G0 ++ G5 ++ (t, A) :: G3) as Hblerg 
      by (autorewrite with step; auto).
     assert (Hwellformed := H0 (G0) (G5++(t,A)::G3) Hblerg G4).
     simpl in Hwellformed.  autorewrite with step in Hwellformed.
     apply Hwellformed.
      auto.

(* TUnboxL. *)
+ apply TUnboxL.
  auto.
  auto.

(* TUnboxP *)
+ apply TUnboxP with (t':=t').
  auto.
  auto.

(* TLam *)
+ apply TLamGen.
  assert (Hbody := H ((th_prog,A) :: G1) G3 eq_refl G2 H2).
   auto.
   auto.

(* TSubType *)
+ apply TSubType.
    auto.

(* TSubtype *)
+ apply TSubtype with (A:= lift A (length G2) (length G1)).
    auto.
    auto.
    auto.

(* TSubtype_L *)
+ apply TSubtype_L with (A:= lift A (length G2) (length G1)).
    auto.
    auto.
    auto.
Qed.

Theorem typing_weakening : forall G1 G2 G3 t a A,
     typing (G1 ++ G3) t a A 
  -> ctx_wf (G2 ++ G3)
  -> typing (lift_c G1 (length G2) 0 ++ G2 ++ G3) t 
            (lift a (length G2) (length G1)) 
            (lift A (length G2) (length G1)).
Proof.
  intros; destruct weakening_combined as [ Htyp _ ]; eauto.
Qed.

Theorem ctx_wf_weakening : forall G1 G2 G3,
     ctx_wf (G1 ++ G3)
  -> ctx_wf (G2 ++ G3)
  -> ctx_wf (lift_c G1 (length G2) 0 ++ G2 ++ G3).
Proof.
  intros; destruct weakening_combined as [ _ Hctx ]; eauto.
Qed.

(***********************************************************)
(******* Automation-friendly versions of weakening *********)
Theorem typing_weakening_pre_app : forall G1 G2 t a A,
     typing G2 t a A
  -> ctx_wf (G1 ++ G2)
  -> typing (G1 ++ G2) t (lift a (length G1) 0)
            (lift A (length G1) 0).
Proof.
  intros;
  apply typing_weakening with (G1 := nil) (G2 := G1) in H; eauto.
Qed.
Hint Resolve typing_weakening_pre_app.

Hint Extern 2 =>
  match goal with
  | [ |- typing (?G1 ++ _) _ (lift _ ?n 0) (lift _ ?n 0) ] =>
    replace n with (length G1) by (simpl; autorewrite with step; auto)
  end.

Theorem typing_weakening_cons : forall G t' B t a A,
     typing G t a A
  -> typing G th_log B exp_star
  -> typing ((t',B) :: G) t (lift a 1 0) (lift A 1 0).
Proof.
  intros G t' B t a A HTyping HTyp_wf;
  apply typing_weakening_pre_app with (G1 := (t',B) :: nil) in HTyping;
  simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons.

Theorem typing_weakening_cons_cons : forall G t'1 B1 t'2 B2 t a A,
     typing G t a A
  -> typing G th_log B1 exp_star
  -> typing ((t'1,B1)::G) th_log B2 exp_star
  -> typing ((t'2,B2)::(t'1,B1)::G) t (lift a 2 0) (lift A 2 0).
Proof.
  intros; match goal with
  | [ H : typing _ _ _ ?A |- typing (?b1 :: ?b2 :: ?G) _ _ (lift ?A _ _) ] =>
    apply typing_weakening_pre_app with (G1 := b1 :: b2 :: nil) in H
  end; simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons_cons.

Theorem star_has_no_type' : forall G t a A,
     typing G t a A
  -> a = exp_star
  -> False.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_star |- _] => inversion H; subst; clear H
  end; eauto.
Qed.  

Theorem star_has_no_type : forall G t A,
   typing G t exp_star A -> False.
Proof.
  intros; eauto using star_has_no_type'.
Qed.

Theorem typing_weakening_app_cons : forall G1 t' B G2 t a A,
     typing G2 t a A
  -> ctx_wf (G1 ++ (t',B) :: G2)
  -> typing (G1 ++ (t',B) :: G2) t (lift a (S (length G1)) 0)
            (lift A (S (length G1)) 0).
Proof.
  intros; invert_ctx_wf;
  match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; inversion H
  | [ Htyp : typing ?G2 _ ?a _ |- typing (?G1' ++ (?t,?B') :: ?G2) _ (lift ?a _ _) _ ] =>
    match B' with
    | exp_star => eapply typing_weakening_cons_star with (t' := t) in Htyp
    | _ => eapply typing_weakening_cons with (t' := t) (B := B') in Htyp
    end; eauto;
    apply typing_weakening_pre_app with (G1 := G1') in Htyp;
    autorewrite with step in *; eauto
  end.
Qed.
Hint Resolve typing_weakening_app_cons.

Theorem typing_weakening_app_cons_closed_typ : forall G1 t' B G2 t a A,
     typing G2 t a A
  -> fvs A = 0
  -> ctx_wf (G1 ++ (t',B) :: G2)
  -> typing (G1 ++ (t',B) :: G2) t (lift a (S (length G1)) 0) A.
Proof.
  intros G1 t' B G2 t a A Htyp HFV HWf;
  replace A with (lift A (S (length G1)) 0) by (autorewrite with step; auto);
  eauto.
Qed.
Hint Resolve typing_weakening_app_cons_closed_typ.
  

(* these three just unfold lift_t, which is getting in the way of automation *)
Theorem typing_weakening_cons_eq : forall G t' B p a b,
     typing G th_log p (exp_eq a b)
  -> typing G th_log B exp_star
  -> typing ((t',B) :: G) th_log (lift p 1 0) 
            (exp_eq (lift a 1 0) (lift b 1 0)).
Proof.
  intros ? ? ? ? ? ? Htyp ?; eapply typing_weakening_cons in Htyp; eauto.
Qed.

Theorem typing_weakening_app_cons_eq : forall G1 G2 t' B p a b,
     typing G2 th_log p (exp_eq a b)
  -> ctx_wf (G1 ++ (t',B) :: G2)
  -> typing (G1 ++ (t',B) :: G2) th_log (lift p (S (length G1)) 0)
         (exp_eq (lift a (S (length G1)) 0) (lift b (S (length G1)) 0)).
Proof.
  intros ? ? ? ? ? ? ? HTyp ?; eapply typing_weakening_app_cons in HTyp;
  simpl in *; eauto.
Qed.

Theorem typing_weakening_cons_cons_eq : forall G tc C tb B p a b,
     typing G th_log p (exp_eq a b)
  -> typing G th_log B exp_star
  -> typing ((tb,B) :: G) th_log C exp_star
  -> typing ((tc,C) :: (tb,B) :: G) th_log (lift p 2 0) 
            (exp_eq (lift a 2 0) (lift b 2 0)).
Proof.
  intros; match goal with
  | [ H : typing _ _ _ (exp_eq _ _) |- typing (?b1 :: ?b2 :: ?G) _ _ _ ] =>
    apply typing_weakening_pre_app with (G1 := b1 :: b2 :: nil) in H
  end; simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons_eq.
Hint Resolve typing_weakening_app_cons_eq.
Hint Resolve typing_weakening_cons_cons_eq.
*)


(******************************************************************)
(* Next, here are some additional lemmas. *)

Theorem open_of_lift : forall A a i,
  {i ~> a} (lift A 1 i) = A.
Proof.
  induction A; intros; simpl; 
  repeat clarify_compares; try omega; rewrite_ih; auto.
Qed.

Theorem weird_lift_lift : forall A i,
  lift (lift A 1 i) 1 i = lift (lift A 1 i) 1 (S i).
Proof.
  induction A; intros; simpl;
   repeat clarify_compares; try omega; rewrite_ih; auto.
Qed.


Lemma better_not_star : forall A A',
     better_typ A A' 
  -> A' = exp_star
  -> False. 
Proof.
  induction 1; 
  try match goal with 
    [ H : FO _ |- _ ] => inversion H; intros
  end;
  try congruence; auto.
Qed.

Lemma star_not_typable : forall G t A B,
  typing G t A B 
  -> A = exp_star 
  -> False. 
Proof.
  induction 1; congruence.
Qed.

Lemma not_star_star : forall G th, 
  typing G th exp_star exp_star
  -> False.
Proof.
  eauto using star_not_typable.
Qed.

Definition typelike (e : exp) : Prop :=
  match e with
    | exp_arr _ _ => True
    | exp_at _ _  => True
    | exp_nat     => True
    | exp_sigma _ _ => True
    | exp_eq _ _  => True
(*
    | exp_Terminates _ _ => True
*)
    | _ => False
  end.

Lemma not_better_star : forall A A',
  better_typ A A'
  -> A' = exp_star
  -> False.
Proof.
  induction 1; intros; 
  try match goal with 
    [ H : FO _ |- _ ] => inversion H; intros
  end;
  try congruence.
Qed.

Lemma not_better_L_star : forall A A',
  better_typ_L A A'
  -> A' = exp_star
  -> False.
Proof.
  induction 1; intros; 
  try match goal with 
    [ H : FO _ |- _ ] => inversion H; intros
  end;
  try congruence; eauto using not_better_star.
Qed.

Lemma type_typelike : forall G th a A,
  typing G th a A
  -> A = exp_star
  -> typelike a.
Proof.
induction 1; intros H_eqA; try rewrite H_eqA in *; simpl in *; 
  auto; congruence.
Qed.

Lemma FO_typelike : forall A,
  FO A -> typelike A.
Proof.
  inversion 1; simpl; auto.
Qed.

Lemma better_typelike : forall A A',
  better_typ A A'
  -> typelike A.
Proof.
  induction 1; simpl; auto using FO_typelike.
Qed.

Lemma better_L_typelike : forall A A',
  better_typ_L A A'
  -> typelike A.
Proof.
  induction 1; simpl; eauto using better_typelike.
Qed.

Lemma term_not_typelike : forall G th a A,
  typing G th a A
  -> typelike A
  -> typelike a
  -> False.
Proof.
induction 1; intros HA Ha;
 simpl in *; auto; eauto using type_typelike.
  
  (* TConv case *)
  destruct A; simpl in *; auto.
  (* A is a variable *)
  destruct (Compare_dec.nat_compare n 0); simpl in *; eauto using type_typelike.
  (* Subtyping case *)
  eauto using better_typelike.
  (* Subtyping_L case *)
  eauto using better_L_typelike.
Qed.
