(* This file is a port of Chris' step_confluence.v *)

Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.
Require Import Program.Equality.

Require Import shared_tactics.
Require Import teqt_defs.
Require Import teqt_db_basics.
Require Import teqt_reduction.

Hint Constructors ptstep.

Theorem open_preserves_pstep : forall a1 a2,
     ptstep a1 a2
  -> forall v1 v2, ptstep v1 v2 -> tval v1
      -> forall n, ptstep ({n ~t> v1}a1) ({n ~t> v2}a2).
Proof.
  induction 1; intros; simpl in *;
  eauto.

  try rewrite open_open_texp_swap with (n2 := 0) by auto;
  try rewrite open_open_texp_swap with (n2 := 0) by auto;
  apply TPCaseZ.
    eauto.

  try rewrite open_open_texp_swap with (n2 := 0) by auto;
  try rewrite open_open_texp_swap with (n2 := 0) by auto.
  apply TPCaseS.
    auto. 
    auto.
    apply IHptstep2.
      rewrite lift_lift_texp_collapse; simpl.
        auto.
        auto.
        auto.
        auto.
(*
  assert (blerg := TPCasePair ({n ~t> v0} v1) ({n ~t> v3} v1')
                              ({n ~t> v0} v2) ({n ~t> v3} v2')
                              (({S (S (S n)) ~t> lift_texp v0 3 0}) b)
                              (({S (S (S n)) ~t> lift_texp v3 3 0}) b')
                              (open_preserves_val _ H _ n H5)
                              (IHptstep1 _ _ H4 H5 n)
                              (open_preserves_val _ H1 _ n H5)
                              (IHptstep2 _ _ H4 H5 n)
                              (IHptstep3 _ _ (lift_preserves_ptstep _ _ H4 _ _) 
                                             (lift_preserves_val _ H5 _ _) (S (S (S n))))). *)
  rewrite open_open_texp_swap by auto.
  rewrite open_open_texp_swap with (a1 := lift_texp v3 1 0) by auto. 
  rewrite open_open_texp_swap with (a1 := lift_texp (lift_texp v3 1 0) 1 0) by auto.
  simpl.
  repeat rewrite lift_lift_texp_collapse by auto.
  simpl.
  change ({S n ~t> lift_texp v3 1 0}) 
    with ({1+n ~t> lift_texp v3 1 0}).
  rewrite open_lift_texp_commute by auto.
  rewrite lift_texp_0_is_id.    
  apply TPCasePair; eauto.


  try rewrite open_open_texp_swap with (n2 := 0) by auto;
  try rewrite open_open_texp_swap with (n2 := 0) by auto.
    extract_open_lift_teqt;
    rewrite open_lift_texp_commute_1 by auto; rewrite lift_texp_0_is_id; simpl;
    rewrite lift_lift_texp_collapse by auto; simpl;
    constructor; autorewrite with step; eauto.

  try rewrite open_open_texp_swap with (n2 := 0) by auto;
  try rewrite open_open_texp_swap with (n2 := 0) by auto.
    extract_open_lift_teqt; simpl;
    constructor; autorewrite with step; eauto.
 
  try rewrite open_open_texp_swap with (n2 := 0) by auto;
  try rewrite open_open_texp_swap with (n2 := 0) by auto.
  simpl.  repeat rewrite lift_lift_texp_collapse by auto.
    simpl.
    replace (S (S (S (S (S n))))) with (3 + (S (S n))) by omega.
    replace 5 with (3 + 2) by omega.
    rewrite open_lift_texp_commute by omega; eauto 7.

 (* um, why didn't auto solve this? *)
 auto.

 auto.
Qed.
Hint Resolve open_preserves_pstep.

Lemma open_preserves_mpstep_l : forall a b1 b2,
     mptstep b1 b2
  -> forall n, mptstep ({n ~t> b1}a) ({n ~t> b2}a).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Immediate open_preserves_mpstep_l.
 
(* Aaron's magic if/then/else *)
Tactic Notation
  "if" tactic(t)
  "then" tactic(t1)
  "else" tactic(t2) :=
  first [ t; first [ t1 | fail 2 ] | t2 ].

(* This builds on if/then/else.  You can use it to demand that two
   metavariables in an Ltac match don't get bound to the same
   thing. *)
Tactic Notation "require_neq" constr(a) constr(b) :=
  if constr_eq a b then fail 2 else idtac.

Lemma open_diamond : forall v1 v2 v b1 b2 b,
  (ptstep b1 b /\ ptstep b2 b)
  -> (ptstep v1 v /\ ptstep v2 v)
  -> tval v1 -> tval v2 
  ->  (ptstep ({0 ~t> v1}b1) ({0 ~t> v}b)) /\ (ptstep ({0 ~t> v2}b2) ({0 ~t> v}b)).
Proof.
intros v1 v2 v b1 b2 b [H1 H2] [H3 H4] H5 H6;
split;
  auto using open_preserves_pstep.
Qed.



Lemma TPLam1_diamond : forall a1 a2 a,
  (ptstep a1 a /\ ptstep a2 a)
 -> ptstep (texp_lam a1) (texp_lam a) /\ ptstep (texp_lam a2) (texp_lam a).
Proof.
  intros a1 a2 a [H1 H2].
  eauto.
Qed.

Lemma TPApp1_diamond : forall a1 a2 a b1 b2 b,
    (ptstep a1 a /\ ptstep a2 a)
 -> (ptstep b1 b /\ ptstep b2 b)
 -> ptstep (texp_app a1 b1) (texp_app a b) 
      /\ ptstep (texp_app a2 b2) (texp_app a b).
Proof.
  intros a1 a2 a b1 b2 b [H1 H2] [H3 H4].
  eauto.
Qed.

Lemma TPNRec1_diamond : forall a1 a2 a,
  (ptstep a1 a /\ ptstep a2 a)
 -> ptstep (texp_nrec a1) (texp_nrec a) /\ ptstep (texp_nrec a2) (texp_nrec a).
Proof.
  intros a1 a2 a [H1 H2].
  eauto.
Qed.

Ltac do_injection := 
  match goal with 
   | [ H : ?F _ = ?F _ |- _ ] => injection H; intros; subst
   | [ H : ?F _ _ = ?F _  _ |- _ ] => injection H; intros; subst
  end.

Lemma inv_ptstepLam_diamond :
  forall a1 a2 c,
      ptstep (texp_lam a1) c
   -> ptstep (texp_lam a2) c
   -> exists c0,
        ptstep a1 c0 /\ ptstep a2 c0.
Proof.
intros a1 a2 c H1 H2;
  inversion H1; inversion H2; intros; subst; do_injection; eauto.
Qed.

Lemma inv_ptstepFun_diamond : 
  forall a1 a2 c,
      ptstep (texp_fun a1) c
   -> ptstep (texp_fun a2) c
   -> exists c0,
        ptstep a1 c0 /\ ptstep a2 c0.
Proof.
intros a1 a2 c H1 H2;
  inversion H1; inversion H2; intros; subst; do_injection; eauto.
Qed.

Lemma inv_ptstepNRec_diamond : 
  forall a1 a2 c,
      ptstep (texp_nrec a1) c
   -> ptstep (texp_nrec a2) c
   -> exists c0,
        ptstep a1 c0 /\ ptstep a2 c0.
Proof.
intros a1 a2 c H1 H2;
  inversion H1; inversion H2; intros; subst; do_injection; eauto.
Qed.


Lemma inv_ptstepSucc_diamond : 
  forall a1 a2 c,
      ptstep (texp_succ a1) c
   -> ptstep (texp_succ a2) c
   -> exists c0,
        ptstep a1 c0 /\ ptstep a2 c0.
Proof.
intros a1 a2 c H1 H2;
  inversion H1; inversion H2; intros; subst; do_injection; eauto.
Qed.


Lemma inv_ptstepPair_diamond : 
  forall a11 a12 a21 a22 c,
      ptstep (texp_pair a11 a12) c
   -> ptstep (texp_pair a21 a22) c
   -> exists c1 c2,
          ptstep a11 c1 /\ ptstep a21 c1
       /\ ptstep a12 c2 /\ ptstep a22 c2.
Proof.
intros a11 a12 a21 a22 c H1 H2.
inversion H1; inversion H2; intros; subst; do_injection; eauto 10.
Qed.

Theorem diamond_pstep : forall a b1,
     ptstep a b1 -> forall b2, ptstep a b2
  -> exists c, ptstep b1 c /\ ptstep b2 c.
Proof.
  (* This tactic does 2 things:
       - Find the IHs and the corresponding (par a b2) derivations and
         apply the IHs appropriately
       - Destruct the resulting existential with some fresh names
         involving "c" *)
  Ltac inst_ptstep := repeat (match goal with
  | [ IHPstep : forall b, ptstep (?P ?a) b -> exists _, ptstep (?P ?b1) _ /\ _,
      HPstep : ptstep ?a ?b |- _ ] =>
    require_neq b1 b;
    assert (ptstep (P a) (P b)) by auto; clear HPstep
  | [ IHPtstep : forall b, ptstep (?P ?a1 ?a2) b -> exists _, ptstep (?P ?b1' ?b2') _ /\ _,
      HPtstep1 : ptstep ?a1 ?b1, HPtstep2 : ptstep ?a2 ?b2 |- _ ] =>
    require_neq b1 b1'; require_neq b2 b2';
    assert (ptstep (P a1 a2) (P b1 b2)) by auto; clear HPtstep1 HPtstep2
  | [ IHPtstep : forall b, ptstep ?a b -> exists _, ptstep ?b1 _ /\ _,
      HPtstep : ptstep ?a ?b |- _ ] =>
    require_neq b1 b; let IH' := fresh "IH" in
    assert (IH' := IHPtstep _ HPtstep); clear IHPtstep;
    let c := fresh "c" in
    let HcL := fresh "HcL" in
    let HcR := fresh "HcR" in
    destruct IH' as [c [HcL HcR]]
  | [ H1 : ptstep (texp_fun _) ?c,
      H2 : ptstep (texp_fun _) ?c |- _] =>
     destruct (inv_ptstepFun_diamond _ _ _ H1 H2) as [? [? ?]]; 
     clear H1; clear H2
  | [ H1 : ptstep (texp_lam _) ?c,
      H2 : ptstep (texp_lam _) ?c |- _] =>
     destruct (inv_ptstepLam_diamond _ _ _ H1 H2) as [? [? ?]]; 
     clear H1; clear H2
  | [ H1 : ptstep (texp_nrec _) ?c,
      H2 : ptstep (texp_nrec _) ?c |- _] =>
     destruct (inv_ptstepNRec_diamond _ _ _ H1 H2) as [? [? ?]]; 
     clear H1; clear H2
  | [ H1 : ptstep (texp_succ _) ?c,
      H2 : ptstep (texp_succ _) ?c |- _] =>
     destruct (inv_ptstepSucc_diamond _ _ _ H1 H2) as [? [? ?]]; 
     clear H1; clear H2
  | [ H1 : ptstep (texp_pair _ _) ?c,
      H2 : ptstep (texp_pair _ _) ?c |- _] =>
     destruct (inv_ptstepPair_diamond _ _ _ _ _ H1 H2) as [? [? [? [? [? ?]]]]]; 
     clear H1; clear H2
  | [ H : ptstep texp_zero _ |- _ ] => inversion H; clear H; subst
  | [ H : ptstep (texp_succ _) _ |- _ ] => inversion H; clear H; subst
  | [ H : ptstep (texp_lam _) _ |- _ ] => inversion H; clear H; subst
  | [ H : ptstep (texp_fun _) _ |- _ ] => inversion H; clear H; subst
  | [ H : ptstep (texp_nrec _) _ |- _ ] => inversion H; clear H; subst
  | [ H : ptstep (texp_succ _) _ |- _ ] => inversion H; clear H; subst
  | [ H : ptstep (texp_pair _ _) _ |- _ ] => inversion H; clear H; subst
  end).

  induction 1; intros; simpl in *; eauto;
  try match goal with
    (* We want to invert the second derivation.   *)
    | [ H : ptstep _ ?b2 |- exists _, _ /\ ptstep ?b2 _ ] =>
      inversion H; subst; simpl in *; clear H
  end.



inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto 10.
(*
inversion H6. subst.
  eauto 10.
destruct (IHptstep2 b2' H9) as [c [Hc1 Hc2]].
clear IHptstep2.
destruct (IHptstep1 v' H0) as [c0 [Hc01 Hc02]].
clear IHptstep1.
  exists ({0 ~t> v'} ({0 ~t> texp_join} c)).
  split.
*)
inst_ptstep; eauto 10.
inst_ptstep; eauto.

inst_ptstep.
eexists;repeat eapply open_diamond; eauto using lift_preserves_ptstep.

{ inst_ptstep.
  + exists (({0 ~t> v1'}) (({0 ~t> lift_texp v2' 1 0}) (({0 ~t> texp_join}) c))).
    split.
      eauto.
      eauto.
  + exists (({0 ~t> c1}) (({0 ~t> lift_texp c0 1 0}) (({0 ~t> texp_join}) c))).
    split.
      repeat apply open_preserves_pstep; eauto.
      eauto.
}

inst_ptstep; eauto.

inst_ptstep;
eexists;
repeat eapply open_diamond;
 eauto 10 using lift_preserves_ptstep.

{ inst_ptstep.
  + exists (({0 ~t> c}) (({0 ~t> texp_fun (lift_texp a' 1 2)}) a')).
    split; eauto.
  + exists (({0 ~t> c}) (({0 ~t> texp_fun (lift_texp c0 1 2)}) c0)).
    split; eauto.
}

inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.

inst_ptstep;
eexists.
repeat eapply open_diamond.
  eauto.
  apply TPLam1_diamond.
  apply TPLam1_diamond.
  apply TPApp1_diamond.
  apply TPNRec1_diamond.
  eauto  using lift_preserves_ptstep.
  eauto.
  eauto.
  eauto.
  eauto.
  eauto.
  eauto.
 
{ inst_ptstep.
  + exists (({0 ~t> c})
              (({0 ~t>
                 texp_lam
                   (texp_lam
                      (texp_app (texp_nrec (lift_texp a' 3 2)) (texp_var 1)))}) a')).
    split; eauto.
  + exists (({0 ~t> c})
              (({0 ~t>
                 texp_lam
                   (texp_lam
                      (texp_app (texp_nrec (lift_texp c0 3 2)) (texp_var 1)))}) c0)).
    split; eauto 10.
}

inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.

{ inst_ptstep;
  exists (({0 ~t> c}) (({0 ~t> texp_fun (lift_texp x 1 2)}) x));
   eauto 10.
}

{ inst_ptstep; exists (({0 ~t> c}) x); eauto. }
   
{ inst_ptstep;
exists (({0 ~t> c})
            (({0 ~t>
              texp_lam
                (texp_lam
                   (texp_app (texp_nrec (lift_texp x 3 2)) (texp_var 1)))})
               x));
  eauto 10.
}

inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.

{ inst_ptstep;
  exists (({0 ~t> x}) (({0 ~t> texp_join}) c));
  eauto. }

inst_ptstep; eauto.
inst_ptstep; eauto.

{ inst_ptstep.
  + exists (({0 ~t> v1'}) (({0 ~t> lift_texp x0 1 0}) (({0 ~t> texp_join}) c)));
    eauto 10.
  + exists (({0 ~t> v1'}) (({0 ~t> lift_texp x0 1 0}) (({0 ~t> texp_join}) c))).
    eauto 10.
  + exists (({0 ~t> a'0}) (({0 ~t> lift_texp x0 1 0}) (({0 ~t> texp_join}) c)));
    eauto 10.
  + exists (({0 ~t> x}) (({0 ~t> lift_texp x0 1 0}) (({0 ~t> texp_join}) c)));
    eauto 10.
}

inst_ptstep; eauto.
inst_ptstep; eauto.
inst_ptstep; eauto.
Qed.

(* Now we extend the diamond property to mpstep, using a "stripping" lemma *)

Theorem stripping_mpstep : forall a b1,
     mptstep a b1 -> forall b2, ptstep a b2
  -> exists c, ptstep b1 c /\ mptstep b2 c.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ HP1 : ptstep ?A ?B1, HMP1 : mptstep ?B1 ?B1',
      HP2 : ptstep ?A ?B2, IH : forall _, ptstep ?B1 _ 
                    -> exists _:_, _ /\ _
     |- exists _:texp, ptstep ?B1' _ /\ mptstep ?B2 _ ] =>
    let B := fresh "B" in
    let C := fresh "C" in
    destruct (diamond_pstep _ _ HP1 _ HP2) as [B [HB1 HB2]]; eauto;
    destruct (IH _ HB1) as [C [HC1 HC2]]; exists C; split; eauto
  end.
Qed.

Theorem diamond_mpstep : forall a b1,
     mptstep a b1 -> forall b2, mptstep a b2
  -> exists c, mptstep b1 c /\ mptstep b2 c.
Proof.
  induction 1; intros; eauto;
  match goal with
  | [ HP1 : ptstep ?A ?A1, HMP1 : mptstep ?A1 ?B1, HMP2 : mptstep ?A ?B2,
      IH : forall _, _ -> exists _:_, _ /\ _ 
     |- exists _:_, mptstep ?B1 _ /\ mptstep ?B2 _ ] =>
    let B2' := fresh B2 "'" in
    destruct (stripping_mpstep _ _ HMP2 _ HP1) as [B2' [HB21 HB22]];
    let C := fresh "C" in
    destruct (IH _ HB22) as [C [? ?]]; eauto
  end.
Qed.

(* Finally, just for good measure, show the diamond property for mstep too *)
Lemma MPTrans : forall a b c,
  mptstep a b -> mptstep b c -> mptstep a c.
Proof.
  induction 1; intros; eauto.
Qed.

Theorem mtstep_mptstep : forall a b,
  mtstep a b -> mptstep a b.
Proof.
  induction 1; eauto.
Qed.
Hint Extern 2 (mptstep ?a ?b) =>
  eapply mtstep_mptstep; eassumption.

Theorem open_preserves_mptstep : forall a1 a2,
     mptstep a1 a2 
  -> forall v1 v2, mptstep v1 v2 -> tval v1 
      -> forall n, mptstep ({n ~t> v1}a1) ({n ~t> v2}a2).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve open_preserves_mptstep.

