(* This is a rough port of Chris' step_db_basics.v to the teqt syntax.
   I basically just changed the relevant identifiers, while keeping
   the proof scripts unchanged. *)

Require Import List.
Require Import Arith.
Require Import Program.Equality.
Require Import Omega.
Require Import NPeano.

Require Import shared_tactics.
Require Import teqt_defs.

Ltac extract_open_lift_teqt :=
  repeat (match goal with
  | [ |- context[ texp_fun (lift_texp ?a 1 2) ] ] =>
    replace (texp_fun (lift_texp a 1 2)) with (lift_texp (texp_fun a) 1 0) in *
         by (simpl; auto)
  | [ |- context[ texp_nrec (lift_texp ?a ?n 2) ] ] =>
    replace (texp_nrec (lift_texp a n 2)) with (lift_texp (texp_nrec a) n 0) in *
         by (simpl; auto)
  | [ |- context[ texp_app (lift_texp (texp_nrec ?a) 1 0) (texp_var 1) ] ] =>
    replace (texp_app (lift_texp (texp_nrec a) 1 0) (texp_var 1))
       with (lift_texp (texp_app (texp_nrec a) (texp_var 0)) 1 0) in *
         by (simpl; auto)
  | [ |- context[ texp_lam (lift_texp ?a 1 1) ] ] =>
    replace (texp_lam (lift_texp a 1 1)) with (lift_texp (texp_lam a) 1 0) in *
         by (simpl; auto)

  | [ |- context[ ttyp_arr ?th ({?n ~T> ?b}?A) ({S ?n ~T> lift_ttyp ?b 1 0}?B)] ] =>
    replace (ttyp_arr th ({n ~T> b}A) ({S n ~T> lift_ttyp b 1 0}B))
       with ({n ~T> b}(ttyp_arr th A B)) by (simpl; auto)
  | [ H : context[ ttyp_arr ?th ({?n ~T> ?b}?A) ({S ?n ~T> lift_ttyp ?b 1 0}?B)] |- _ ] =>
    replace (ttyp_arr th ({n ~T> b}A) ({S n ~T> lift_ttyp b 1 0}B))
       with ({n ~T> b}(ttyp_arr th A B)) in H by (simpl; auto)
  | [ H : context[ ttyp_arr ?th (lift_ttyp ?A 1 0) (lift_ttyp ?B 1 1) ] |- _ ] =>
    replace (ttyp_arr th (lift_ttyp A 1 0) (lift_ttyp B 1 1))
       with (lift_ttyp (ttyp_arr th A B) 1 0) in H by (simpl; auto)
  | [ |- context[ ttyp_arr ?th (lift_ttyp ?A ?n ?m) (lift_ttyp ?B ?n (S ?m))] ] =>
    replace (ttyp_arr th (lift_ttyp A n m) (lift_ttyp B n (S m)))
       with (lift_ttyp (ttyp_arr th A B) n m) by (simpl; auto)
  | [ H : context[ ttyp_arr ?th (lift_ttyp ?A ?n ?m)  (lift_ttyp ?B ?n (S ?m))] |- _ ] =>
    replace (ttyp_arr th (lift_ttyp A n m) (lift_ttyp B n (S m)))
       with (lift_ttyp (ttyp_arr th A B) n m) in H by (simpl; auto)
  | [ |- context[ ttyp_sigma ({?n ~T> ?b}?A) ({S ?n ~T> lift_ttyp ?b 1 0}?B)] ] =>
     replace (ttyp_sigma ({n ~T> b}A) ({S n ~T> lift_ttyp b 1 0}B))
        with ({n ~T> b}(ttyp_sigma A B)) by (simpl; auto)
  | [ H : context[ ttyp_sigma ({?n ~T> ?b}?A) ({S ?n ~T> lift_ttyp ?b 1 0}?B)] |- _ ] =>
     replace (ttyp_sigma ({n ~T> b}A) ({S n ~T> lift_ttyp b 1 0}B))
        with ({n ~T> b}(ttyp_sigma A B)) in H by (simpl; auto)
  | [ |- context [ttyp_sigma (lift_ttyp ?A1 ?n ?m) (lift_ttyp ?A2 ?n (S ?m))] ] =>
      replace (ttyp_sigma (lift_ttyp A1 n m) (lift_ttyp A2 n (S m))) with
              (lift_ttyp (ttyp_sigma A1 A2) n m) by auto
  | [ H : context [ttyp_sigma (lift_ttyp ?A1 ?n ?m) (lift_ttyp ?A2 ?n (S ?m))] |- _ ] =>
      replace (ttyp_sigma (lift_ttyp A1 n m) (lift_ttyp A2 n (S m))) with
              (lift_ttyp (ttyp_sigma A1 A2) n m) in H by auto
  | [ |- context [ ttyp_eq ({?n ~t> ?b}?a1) ({?n ~t> ?b}?a2) ] ] =>
    replace (ttyp_eq ({n ~T> b}a1) ({n ~T> b}a2))
       with ({n ~T> b}(ttyp_eq a1 a2)) by (simpl; auto)
  | [ H : context [ ttyp_eq ({?n ~t> ?b}?a1) ({?n ~t> ?b}?a2) ] |- _ ] =>
    replace (ttyp_eq ({n ~t> b}a1) ({n ~t> b}a2))
       with ({n ~t> b}(ttyp_eq a1 a2)) in H by (simpl; auto)
  end).

Ltac destruct_typ_eq :=
  match goal with
  | [ H : {_ ~T> _}?A = ttyp_nat |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~T> _}?A = ttyp_arr _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~T> _}?A = ttyp_eq _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~T> _}?A = ttyp_sigma _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
(*
  | [ H : {_ ~T> _}?A = ttyp_terminates _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
*)
  end.

Hint Rewrite <- plus_n_O : step.
Hint Rewrite app_length : step.


Ltac val_inversion :=
  repeat (match goal with
  | [ HV : tval (texp_succ _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : tval (texp_ncase _ _ _) |- _ ] => inversion HV
  | [ HV : tval (texp_pcase _ _) |- _ ] => inversion HV
  | [ HV : tval (texp_pair _ _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : tval (texp_app _ _) |- _ ] => inversion HV
  end).

(* Opening and lifting closed terms/types does nothing *)
Theorem open_closed_texp_id_gen : forall a n b,
  fvs_texp a <= n -> {n ~t> b}a = a.
Proof.
  induction a; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem open_closed_texp_id : forall a n b,
  fvs_texp a = 0 -> {n ~t> b}a = a.
Proof.
  intros; eauto using open_closed_texp_id_gen.
Qed.
Hint Resolve open_closed_texp_id open_closed_texp_id_gen.
Hint Rewrite open_closed_texp_id_gen using (simpl; eauto 2; fail) : step.

Theorem open_closed_ttyp_id_gen : forall A n b,
  fvs_ttyp A <= n -> {n ~T> b}A = A.
Proof.
  induction A; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem open_closed_ttyp_id : forall A n b,
  fvs_ttyp A = 0 -> {n ~T> b}A = A.
Proof.
  intros; eauto using open_closed_ttyp_id_gen.
Qed.
Hint Resolve open_closed_ttyp_id open_closed_ttyp_id_gen.
Hint Rewrite open_closed_ttyp_id_gen using (simpl; eauto 2; fail) : step.

Theorem lift_closed_texp_id_gen : forall a n m,
  fvs_texp a <= m -> lift_texp a n m = a.
Proof.
  induction a; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem lift_closed_texp_id : forall a n m,
  fvs_texp a = 0 -> lift_texp a n m = a.
Proof.
  intros; eauto using lift_closed_texp_id_gen.
Qed.
Hint Rewrite lift_closed_texp_id using (eauto 3; fail) : step.
Hint Rewrite lift_closed_texp_id_gen using (eauto 3; fail) : step.

Theorem lift_closed_ttyp_id_gen : forall A n m,
  fvs_ttyp A <= m -> lift_ttyp A n m = A.
Proof.
  induction A; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem lift_closed_ttyp_id : forall A n m,
  fvs_ttyp A = 0 -> lift_ttyp A n m = A.
Proof.
  intros; eauto using lift_closed_ttyp_id_gen.
Qed.
Hint Rewrite lift_closed_ttyp_id using (eauto 3; fail) : step.


Theorem lift_texp_0_is_id : forall (b : texp) (m : nat),
  lift_texp b 0 m = b.
Proof.
  induction b; intros; simpl in *; rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite lift_texp_0_is_id : step.

Theorem lift_ttyp_0_is_id : forall (B : ttyp) (m : nat),
  lift_ttyp B 0 m = B.
Proof.
  induction B; intros; simpl in *; rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite lift_ttyp_0_is_id : step.

(* Swapping and commuting lemmas *)
Theorem lift_lift_texp_swap_gen : forall b m1 m2 n1 n2,
    m2 <= m1
  ->   lift_texp (lift_texp b n1 m1) n2 m2
     = lift_texp (lift_texp b n2 m2) n1 (m1 + n2).
Proof.
  induction b; intros; simpl; eauto; rewrite_ih; auto;
  db_var_case.
Qed.

Theorem lift_lift_texp_swap_0 : forall b m1 n1 n2,
       lift_texp (lift_texp b n1 m1) n2 0
     = lift_texp (lift_texp b n2 0) n1 (n2 + m1).
Proof.
  intros; subst; rewrite lift_lift_texp_swap_gen; eauto with arith.
Qed.
Hint Rewrite <- lift_lift_texp_swap_0 : step.
Hint Extern 4 =>
  match goal with
  | [ H : context [ lift_texp (lift_texp ?A 1 1) ?n (S ?m) ]
     |- context [lift_texp (lift_texp ?A ?n ?m) 1 1] ] =>
    rewrite lift_lift_texp_swap_gen with (b := A) by omega
  | [ |- context[ lift_texp (texp_fun (lift_texp _ 1 2)) _ (S ?m) ] ] =>
    extract_open_lift_teqt; replace (S m) with (1 + m) by auto;
    rewrite <- lift_lift_texp_swap_0
  | [ |- context[ lift_texp (texp_nrec (lift_texp _ 1 2)) _ (S ?m) ] ] =>
    extract_open_lift_teqt; replace (S m) with (1 + m) by auto;
    rewrite <- lift_lift_texp_swap_0
  | [ |- context[lift_texp (lift_texp _ _ ?m1) ?n2 0] ] =>
    rewrite lift_lift_texp_swap_0
  | [ H : context[lift_texp (lift_texp _ _ ?m1) ?n2 0] |- _ ] =>
    rewrite lift_lift_texp_swap_0 in H
  end.

Theorem lift_lift_texp_collapse : forall b n1 n2 m1 m2,
     m1 <= m2
  -> m2 <= m1 + n1
  ->   lift_texp (lift_texp b n1 m1) n2 m2
     = lift_texp b (n1 + n2) m1.
Proof.
  induction b; intros; simpl in *; auto; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_lift_texp_collapse using omega : step.

(* Special case of the last theorem that comes up *)
Theorem lift_lift_texp_collapse_1 : forall b n1,
       lift_texp (lift_texp b n1 0) 1 0
     = lift_texp b (S n1) 0.
Proof.
  intros; autorewrite with step; simpl; auto.
Qed.
Hint Rewrite lift_lift_texp_collapse_1 : step.

Theorem lift_lift_ttyp_swap_gen : forall b m1 m2 n1 n2,
    m2 <= m1
  ->   lift_ttyp (lift_ttyp b n1 m1) n2 m2
     = lift_ttyp (lift_ttyp b n2 m2) n1 (m1 + n2).
Proof.
  induction b; intros; simpl; eauto; rewrite_ih;  auto;
  db_var_case;
  try (f_equal; rewrite lift_lift_texp_swap_gen; now auto).
Qed.    

Theorem lift_lift_ttyp_swap_0 : forall b m1 n1 n2,
       lift_ttyp (lift_ttyp b n1 m1) n2 0
     = lift_ttyp (lift_ttyp b n2 0) n1 (n2 + m1).
Proof.
  intros; subst; rewrite lift_lift_ttyp_swap_gen; eauto with arith.
Qed.
Hint Rewrite <- lift_lift_ttyp_swap_0 : step.
Hint Extern 4 =>
  match goal with
  | [ H : context [ lift_ttyp (lift_ttyp ?A 1 1) ?n (S ?m) ]
     |- context [lift_ttyp (lift_ttyp ?A ?n ?m) 1 1] ] =>
    rewrite lift_lift_ttyp_swap_gen with (b := A) by omega
  | [ |- context[lift_ttyp (lift_ttyp _ _ ?m1) ?n2 0] ] =>
    rewrite lift_lift_ttyp_swap_0
  | [ H : context[lift_ttyp (lift_ttyp _ _ ?m1) ?n2 0] |- _ ] =>
    rewrite lift_lift_ttyp_swap_0 in H
  end.

Theorem lift_lift_ttyp_collapse : forall b n1 n2 m1 m2,
     m1 <= m2
  -> m2 <= m1 + n1
  ->   lift_ttyp (lift_ttyp b n1 m1) n2 m2
     = lift_ttyp b (n1 + n2) m1.
Proof.
  induction b; intros; simpl in *; auto; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_lift_ttyp_collapse using omega : step.

Theorem lift_lift_ttyp_collapse_1 : forall b n1,
       lift_ttyp (lift_ttyp b n1 0) 1 0
     = lift_ttyp b (S n1) 0.
Proof.
  intros; autorewrite with step; simpl; auto.
Qed.
Hint Rewrite lift_lift_ttyp_collapse_1 : step.

Theorem lift_open_texp_commute : forall B b i n m,
     i <= m
  ->   lift_texp ({i ~t> b}B) n m
     = {i ~t> lift_texp b n m}(lift_texp B n (S m)).
Proof.
  induction B; intros; simpl in *; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_open_texp_commute using omega : step.
Hint Extern 4 =>
  match goal with
  | [ |- context [lift_texp ({_ ~t> _}_) _ _] ] =>
    rewrite lift_open_texp_commute by (auto with arith)
  end.

Theorem open_lift_texp_commute : forall b a n1 n2 i m,
     m <= n2
  -> m <= i 
  -> {n1 + i ~t> (lift_texp a (n1 + n2) 0)} (lift_texp b n1 m) 
   = lift_texp ({i ~t> lift_texp a n2 0} b) n1 m.
Proof.
  induction b; intros; simpl in *; eauto;
  autorewrite with step; simpl in *;
  repeat (match goal with
  | [ |- context [ S (_ + _) ] ] =>
    repeat (rewrite plus_n_Sm)
  | [ |- context [ (_ + _ + _) ] ] =>
    repeat (rewrite <- plus_assoc)
  end); rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite <- open_lift_texp_commute using omega : step.

Theorem open_lift_texp_commute_1 : forall b a n2 i m,
     m <= n2 -> m <= i
  ->   {S i ~t> (lift_texp a (S n2) 0)}(lift_texp b 1 m)
     = lift_texp ({i ~t> lift_texp a n2 0} b) 1 m.
Proof.
  intros b a n2 i m ? ?; replace (S i) with (1 + i) by auto;
  autorewrite with step in *; eauto.
Qed.
Hint Rewrite <- open_lift_texp_commute_1 using omega : step.

Theorem open_lift_texp_commute_0 : forall b a n1,
    {n1 ~t> lift_texp a n1 0} (lift_texp b n1 0)
  = lift_texp ({0 ~t> a} b) n1 0.
Proof.
  intros b a n1; replace ({0 ~t> a} b) with ({0 ~t> lift_texp a 0 0} b);
  try rewrite <- open_lift_texp_commute by omega;
  autorewrite with step; eauto.
Qed.

Theorem open_lift_texp_commute_1_closed : forall b a i m,
     m <= i
  -> fvs_texp a = 0
  ->   {S i ~t> a} (lift_texp b 1 m) 
     = lift_texp ({i ~t> a} b) 1 m.
Proof.
  intros b a i m H1 HFV;
  replace ({i ~t> a}) with ({i ~t> lift_texp a m 0})
       by (autorewrite with step; eauto);
  replace ({S i ~t> a}) with ({S i ~t> lift_texp a (S m) 0})
       by (autorewrite with step; eauto);
  autorewrite with step; eauto.
Qed.
Hint Rewrite <- open_lift_texp_commute_1_closed using (simpl; omega): step.


Theorem lift_open_ttyp_commute : forall B b i n m,
     i <= m
  ->   lift_ttyp ({i ~T> b}B) n m
     = {i ~T> lift_texp b n m}(lift_ttyp B n (S m)).
Proof.
  induction B; intros; simpl in *; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_open_ttyp_commute using omega : step.
Hint Extern 4 =>
  match goal with
  | [ |- context [lift_ttyp ({_ ~T> _}_) _ _] ] =>
    rewrite lift_open_ttyp_commute by (auto with arith)
  end.

Theorem open_lift_ttyp_commute : forall b a n1 n2 i m,
     m <= n2
  -> m <= i 
  -> {n1 + i ~T> (lift_texp a (n1 + n2) 0)} (lift_ttyp b n1 m) 
   = lift_ttyp ({i ~T> lift_texp a n2 0} b) n1 m.
Proof.
  induction b; intros; simpl in *; eauto;
  autorewrite with step; simpl in *;
  repeat (match goal with
  | [ |- context [ S (_ + _) ] ] =>
    repeat (rewrite plus_n_Sm)
  | [ |- context [ (_ + _ + _) ] ] =>
    repeat (rewrite <- plus_assoc)
  end); rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite <- open_lift_ttyp_commute using omega : step.

Theorem open_lift_ttyp_commute_1 : forall b a n2 i m,
     m <= n2 -> m <= i
  ->   {S i ~T> (lift_texp a (S n2) 0)}(lift_ttyp b 1 m)
     = lift_ttyp ({i ~T> lift_texp a n2 0} b) 1 m.
Proof.
  intros b a n2 i m ? ?; replace (S i) with (1 + i) by auto;
  autorewrite with step in *; eauto.
Qed.
Hint Rewrite <- open_lift_ttyp_commute_1 using omega : step.

Theorem open_lift_ttyp_commute_0 : forall b a n1,
    {n1 ~T> lift_texp a n1 0} (lift_ttyp b n1 0)
  = lift_ttyp ({0 ~T> a} b) n1 0.
Proof.
  intros b a n1; replace ({0 ~T> a} b) with ({0 ~T> lift_texp a 0 0} b);
  try rewrite <- open_lift_commute by omega;
  autorewrite with step; eauto.
Qed.

Theorem open_lift_ttyp_commute_1_closed : forall b a i m,
     m <= i
  -> fvs_texp a = 0
  ->   {S i ~T> a} (lift_ttyp b 1 m) 
     = lift_ttyp ({i ~T> a} b) 1 m.
Proof.
  intros b a i m H1 HFV;
  replace ({i ~T> a}) with ({i ~T> lift_texp a m 0})
       by (autorewrite with step; eauto);
  replace ({S i ~T> a}) with ({S i ~T> lift_texp a (S m) 0})
       by (autorewrite with step; eauto);
  autorewrite with step; eauto.
Qed.
Hint Rewrite <- open_lift_ttyp_commute_1_closed using (simpl; omega): step.


Theorem lift_ctx_length : forall n m G,
  length G = length (lift_c G n m).
Proof.
  induction G ; intros; simpl in *; eauto.
Qed.
Hint Rewrite <- lift_ctx_length : step.
Hint Resolve lift_ctx_length.

Theorem lift_ctx_app : forall G1 G2 n m,
    lift_c (G1 ++ G2) n m 
  = (lift_c G1 n (m + length G2)) ++ lift_c G2 n m.
Proof.
  induction G1; intros; simpl in *; eauto;
  try rewrite app_length; repeat (f_equal; try omega; eauto).
Qed.
Hint Rewrite lift_ctx_app : step.


Theorem open_texp_under_lift : forall b a i n m,
     i <= n + m
  -> m <= i
  -> {i ~t> a} (lift_texp b (S n) m) = lift_texp b n m.
Proof.
  induction b; intros; simpl in *; rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite open_texp_under_lift using omega : step.

Theorem open_texp_into_lift : forall A n,
  {n ~t> texp_var (S n)} (lift_texp A 2 (S n)) = lift_texp A 1 n.
Proof.
  induction A; intros; simpl; eauto;
  try solve [autorewrite with step; repeat rewrite_ih; eauto];
  db_var_case.
Qed.

Theorem open_open_texp_swap : forall b n1 n2 a1 a2,
    n2 <= n1
  ->   {n1 ~t> a1}({n2 ~t> a2}b)
     = {n2 ~t> ({n1 ~t> a1}a2)}({S n1 ~t> lift_texp a1 1 n2}b).
Proof.
  induction b; intros; simpl in *; rewrite_ih; eauto;
  try match goal with
  | [ |- _ ] =>
    (* a little bit of fanciness so that we can see open_lift_commute applies *)
    f_equal; auto; match goal with
    | [ |- context[lift_texp ({?i ~t> ?a}?b) ?n 0] ] =>
      replace (lift_texp ({i ~t> a}b) n 0) with (lift_texp ({i ~t> lift_texp a 0 0}b) n 0)
        by (rewrite lift_texp_0_is_id; eauto)
    end; rewrite <- open_lift_texp_commute; eauto
  end;
  try db_var_case.
Qed.

Theorem open_ttyp_under_lift : forall b a i n m,
     i <= n + m
  -> m <= i
  -> {i ~T> a} (lift_ttyp b (S n) m) = lift_ttyp b n m.
Proof.
  induction b; intros; simpl in *; rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite open_ttyp_under_lift using omega : step.

Theorem open_ttyp_into_lift : forall A n,
  {n ~T> texp_var (S n)} (lift_ttyp A 2 (S n)) = lift_ttyp A 1 n.
Proof.
  induction A; intros; simpl; eauto;
  try solve [autorewrite with step; repeat rewrite_ih; eauto];
  db_var_case;
  try (f_equal; (rewrite open_texp_into_lift || rewrite_ih); now auto).
Qed.


Theorem open_open_ttyp_swap : forall b n1 n2 a1 a2,
    n2 <= n1
  ->   {n1 ~T> a1}({n2 ~T> a2}b)
     = {n2 ~T> ({n1 ~t> a1}a2)}({S n1 ~T> lift_texp a1 1 n2}b).
Proof.
  induction b; intros; simpl in *; rewrite_ih; eauto;

f_equal;
  try match goal with
    | [ |- context[lift_texp ({?i ~t> ?a}?b) ?n 0] ] => 
     replace (lift_texp ({i ~t> a}b) n 0) with (lift_texp ({i ~t> lift_texp a 0 0}b) n 0) by (rewrite lift_texp_0_is_id; eauto)
  end;
  autorewrite with step;
  try rewrite open_open_texp_swap; 
  auto.
Qed.

Lemma open_lift_texp_id : forall a b m n,
  ({n ~t> b} (lift_texp a (S m) n)) = lift_texp a m n.
Proof.
  induction a; simpl; intros; auto; 
   try rewrite IHa;
   try rewrite IHa1;
   try rewrite IHa2;
   try rewrite IHa3;
   eauto.
    db_var_case.
Qed.

Lemma open_lift_ttyp_id : forall A b m n,
  ({n ~T> b} (lift_ttyp A (S m) n)) = lift_ttyp A m n.
Proof.
  induction A; simpl; intros; auto; 
   try rewrite IHA;
   try rewrite IHA1;
   try rewrite IHA2;
   eauto;
   repeat rewrite open_lift_texp_id; auto.
Qed.  


(* The de bruijn operations meet reduction *)
Theorem lift_preserves_val : forall v,
  tval v -> forall n m, tval (lift_texp v n m).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve lift_preserves_val.

Theorem open_preserves_val : forall v,
  tval v -> forall v' n, tval v' -> tval ({n ~t> v'} v).
Proof.
  induction 1; intros; simpl in *; eauto;
  clarify_compares; simpl; eauto.
Qed.
Hint Resolve open_preserves_val.

Theorem ptstep_preserves_val : forall v v',
  ptstep v v' -> tval v -> tval v'.
Proof.
  induction 1; intros; subst; val_inversion; eauto.
Qed.
Hint Immediate ptstep_preserves_val.

Hint Constructors ptstep.

Theorem lift_preserves_ptstep : forall a b,
     ptstep a b
  -> forall n m, 
       ptstep (lift_texp a n m) (lift_texp b n m).
Proof.  
  induction 1; intros; simpl in *; eauto;
  autorewrite with step;
  eauto;
match goal with 
 | [ |- ptstep (texp_app (texp_fun _) _) _ ] =>
   repeat (rewrite lift_open_texp_commute by auto); eauto;
          extract_open_lift_teqt; replace (S m) with (1 + m) by (simpl; auto);
          rewrite <- lift_lift_texp_swap_0; simpl; auto
  | [ |- ptstep (texp_app (texp_nrec _) _) _ ] =>
    repeat (rewrite lift_open_texp_commute by auto); simpl;
    replace (S (S (S (S (S m))))) with (S (S m) + 3) by omega;
    rewrite <- lift_lift_texp_swap_gen by omega; eauto
  end.
Qed.
Hint Resolve lift_preserves_ptstep.

Hint Constructors mptstep.

Theorem lift_preserves_mptstep : forall a b,
    mptstep a b
 -> forall n m,
      mptstep (lift_texp a n m) (lift_texp b n m).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve lift_preserves_mptstep.

Theorem open_preserves_ptstep_r :  forall a a',
  ptstep a a' -> forall v n, tval v -> ptstep ({n ~t> v}a) ({n ~t> v}a').
Proof.
  induction 1; intros; simpl in *; eauto 7;
  rewrite open_open_texp_swap with (n2 := 0) by omega; eauto;
  try (rewrite open_open_texp_swap with (n2 := 0) by omega; eauto);
  simpl; autorewrite with step; simpl;
  try rewrite open_lift_texp_commute_1 by omega; eauto 6.

(* The PCasePair case, sorry for not automating more. *)
rewrite open_open_texp_swap with  (a2 := texp_join) (n2:=0) by omega;
rewrite lift_texp_0_is_id; 
rewrite open_closed_texp_id with (a := texp_join) by auto;
rewrite lift_lift_texp_collapse by auto; auto.

(* natrec *)
  replace (S (S (S (S (S n))))) with (3 + (S (S n))) by omega.
  replace 5 with (3 + 2) by omega.
  rewrite open_lift_texp_commute by omega.  eauto 6.
Qed.
Hint Resolve open_preserves_ptstep_r.

Theorem open_preserves_mptstep_r : forall a a',
  mptstep a a' -> forall v n, tval v -> mptstep ({n ~t> v}a) ({n ~t> v}a').
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve open_preserves_mptstep_r.

Theorem open_preserves_ptstep_l : forall b a a',
  ptstep a a' -> forall n, ptstep ({n ~t> a}b) ({n ~t> a'}b).
Proof.
  induction b; intros; simpl in *; eauto 7;
  db_var_case.
Qed.
Hint Resolve open_preserves_ptstep_l.
