Require Import Arith.
Require Import Bool.
Require Import List.

Require Import shared_tactics.
Require Import shared_defs.
Require Import teqt_defs.

Inductive pre_tctx_wf : tctx -> Prop :=
 | TCWFNil  : pre_tctx_wf nil
 | TCWFTrm : forall A G,
       pre_tkinding G A 
    -> pre_tctx_wf G
    -> pre_tctx_wf (A :: G)

with pre_ttyping : tctx -> theta -> texp -> ttyp -> Prop :=
 | PTTVar : forall (G1 G3:tctx) (th:theta) (n:nat) (A:ttyp),
     tctx_wf (G1 ++ A :: G3) ->
     pre_tkinding (G1 ++ A :: G3) (lift_ttyp A (S n) 0) ->
     n = length G1 ->
     pre_ttyping (G1 ++ A :: G3) th (texp_var n) (lift_ttyp A (S n) 0)
 | PTTApp : forall (G:tctx) (th1 th2:theta) (a b : texp) (A B:ttyp),
       pre_ttyping G th1 a (ttyp_arr th2 A B)
    -> pre_ttyping G th1 b A
    -> better_theta th2 th1
    -> pre_tkinding G  ({0 ~T> b} B)
    -> pre_ttyping G th1 (texp_app a b) ({0 ~T> b} B)
 | PTTLam : forall (G:tctx) th1 th2 (b :texp) (A B:ttyp),
       pre_ttyping (A :: G) th2 b B
    -> pre_tkinding G (ttyp_arr th2 A B)
    -> pre_ttyping G th1 (texp_lam b) (ttyp_arr th2 A B)
 | PTTRec : forall (G:tctx) th (a :texp) (A B:ttyp),
     pre_ttyping ((lift_ttyp (ttyp_arr th_prog A B) 1 0) :: A :: G)
            th_prog a (lift_ttyp B 1 0) ->
     pre_tkinding G (ttyp_arr th_prog A B) ->
     pre_ttyping G th (texp_fun a) (ttyp_arr th_prog A B)
 | PTTNRec : forall (G:tctx) th (a :texp)  (B:ttyp),
     pre_ttyping (   (ttyp_arr th_log ttyp_nat 
                     (ttyp_arr th_log (ttyp_eq (texp_succ (texp_var 0))
                                               (texp_var 1))
                              (lift_ttyp (lift_ttyp B 1 1) 1 0)))
             :: ttyp_nat :: G)
            th_log a (lift_ttyp B 1 0) ->
     pre_tkinding G (ttyp_arr th_log ttyp_nat B) ->
     pre_ttyping  G th (texp_nrec a) (ttyp_arr th_log ttyp_nat B)
 | PTTZero : forall (G:tctx) (th:theta),
     tctx_wf G ->
     pre_ttyping G th texp_zero ttyp_nat
 | PTTSuc : forall (G:tctx) (th:theta) (b:texp),
     pre_ttyping G th b ttyp_nat ->
     pre_ttyping G th (texp_succ b) ttyp_nat
 | PTTPair : forall G th A B a b,
      pre_tkinding G (ttyp_sigma A B)
   -> pre_ttyping  G th a A
   -> pre_tkinding G ({0 ~T> a}B)
   -> pre_ttyping  G th b ({0~T>a}B)
   -> pre_ttyping  G th (texp_pair a b) (ttyp_sigma A B)

 | PTTPCase : forall G t A1 A2 B a b,
      pre_ttyping G t a (ttyp_sigma A1 A2)
   -> pre_ttyping (ttyp_eq (texp_pair (texp_var 1) (texp_var 0)) (lift_texp a 2 0)
              :: A2 
              :: A1
              :: G)
             t b (lift_ttyp B 3 0)
   -> pre_tkinding G B
   -> pre_ttyping G t (texp_pcase a b) B

 | PTTNCase : forall (G:tctx) (t:theta) (a b1 b2 :texp) (A:ttyp),
     pre_ttyping G t a ttyp_nat ->
     pre_ttyping ((ttyp_eq texp_zero a) :: G) t b1 (lift_ttyp A 1 0) ->
     pre_ttyping ((ttyp_eq (texp_succ (texp_var 0)) (lift_texp a 1 0)) 
               :: ttyp_nat :: G)
             t b2 (lift_ttyp A 2 0) ->
     pre_tkinding G A ->
     pre_ttyping G t (texp_ncase a b1 b2) A 
 | PTTJoin : forall G a b c t1 t2 th A B,
       mptstep a c -> mptstep b c
    -> pre_ttyping G t1 a A -> pre_ttyping G t2 b B
    -> pre_tkinding G A -> pre_tkinding G B
    -> pre_ttyping G th texp_join (ttyp_eq a b)
 | PTTConv : forall G p a b1 b2  t A,
       pre_ttyping G th_log p (ttyp_eq b1 b2)
(* These two lines are the difference between ttyping and pre_ttyping. *)
(*    -> pre_ttyping G th_prog b1 B1 -> pre_tkinding G B1
      -> pre_ttyping G th_prog b2 B2 -> pre_tkinding G B2 *)
    -> pre_ttyping G t a ({0 ~T> b1}A)
    -> pre_tkinding G ({0 ~T> b2}A)
    -> pre_ttyping G t a ({0 ~T> b2}A)
 | PTTSubtype : forall G th a A A',
       pre_ttyping G th a A
    -> better_ttyp A A'
    -> pre_tkinding G A'
    -> pre_ttyping G th a A'

with pre_tkinding : tctx -> ttyp -> Prop :=
 | TKArr : forall G A th B,
       pre_tkinding G A 
    -> pre_tkinding (A :: G) B
    -> pre_tkinding G (ttyp_arr th A B)
 | TKSigma : forall G A B,
       pre_tkinding G A 
    -> pre_tkinding (A :: G) B
    -> pre_tkinding G (ttyp_sigma A B)
 | TKNat : forall G,
     tctx_wf G -> pre_tkinding G ttyp_nat
 | TKEq  : forall G a b A t1 B t2,
       pre_ttyping G t1 a A -> pre_ttyping G t2 b B
    -> pre_tkinding G A -> pre_tkinding G B
    -> pre_tkinding G (ttyp_eq a b)
 .


Hint Constructors pre_tctx_wf pre_ttyping pre_tkinding.
Hint Resolve subsume_theta.

Scheme pre_ttyping_ind_mut  := Induction for pre_ttyping Sort Prop
  with pre_tkinding_ind_mut := Induction for pre_tkinding Sort Prop                        
  with pre_tctx_wf_ind_mut  := Induction for pre_tctx_wf Sort Prop.



Theorem prettyping_2_ttyping :
  forall G t a A,
    pre_ttyping G t a A -> ttyping G t a A.
Proof.
  apply (pre_ttyping_ind_mut 
         (fun G th a A _ =>
            ttyping G th a A /\ tkinding G A)
         (fun G A _ =>
            tkinding G A)
         (fun G _ => tctx_wf G));
  intros;
 simpl in *; try apply conj; intros; 
  repeat match goal with 
   | [ H : ttyping _ _ _ _ /\ tkinding _ _ |- _] => destruct H
  end;
 auto; eauto.

(* TPTConv. This is where we need regularity. *)
inversion H3.
subst.
apply TTConv with (p:=p) (b1:=b1) (b2:=b2) (B1 := A0) (B2 := B); eauto.
Qed.


Theorem ttyping_2_prettyping :
  forall G t a A,
    ttyping G t a A -> pre_ttyping G t a A.
Proof.
  apply (ttyping_ind_mut 
         (fun G th a A _ =>
            pre_ttyping G th a A)
         (fun G A _ =>
            pre_tkinding G A)
         (fun G _ => pre_tctx_wf G));
  intros;
 simpl in *; intros;
 auto; eauto.
Qed.

Corollary ttyping_iff_prettyping :
  forall G t a A,
     ttyping G t a A <-> pre_ttyping G t a A.
Proof.
  split.
    apply ttyping_2_prettyping.
 
    intros H. apply prettyping_2_ttyping in H. assumption.
Qed.