(* This file is a staight port of Chris' step_reduction_basics.v,
   only changing the names of identifiers.  *)

Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.

Require Import teqt_defs.
Require Import teqt_db_basics.
Require Import shared_tactics.

(********************)

Hint Constructors tstep.
Hint Constructors ptstep.
Hint Constructors mtstep.
Hint Constructors mptstep.

(* Some facts about values *)
Theorem values_dont_step : forall a,
  tval a -> forall b, tstep a b -> False.
Proof.
  induction 1; intros ? HStep;
  inversion HStep; subst; eauto.
Qed.

Theorem val_mtstep : forall a b,
  tval a -> mtstep a b ->  a = b.
Proof.
  inversion 2; eauto; subst;
  edestruct values_dont_step; eauto.
Qed.

Ltac step_val_elim :=
  match goal with
  | [ HS : tstep ?a _ |- _ ] =>
    let HVal := fresh "HVal" in assert (tval a) as HVal by eauto;
    destruct (values_dont_step _ HVal _ HS)
  | [ HS : mtstep _ ?a _ |- _ ] =>
    let HVal := fresh "HVal" in assert (tval a) as HVal by eauto;
    destruct (val_mtstep _ _ _ HVal HS); subst
  end.

Theorem tstep_ptstep : forall a b,
  tstep a b -> ptstep a b.
Proof.
  induction 1; eauto.
Qed.
Hint Immediate tstep_ptstep. 

Theorem mtstep_mptstep : forall a b,
  mtstep a b -> mptstep a b.
Proof.
  induction 1.
    auto.
    eauto.
Qed.

Lemma tstep_mtstep : forall a b, 
  tstep a b -> mtstep a b.
Proof.
  intros.
  apply MSTStep with (b:=b); auto.
Qed.

(* Congruence rules for MS *)

Theorem MSTrans : forall a b c,
  mtstep a b -> mtstep b c -> mtstep  a c.
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSApp1 : forall a a' b,
  mtstep a a' -> mtstep (texp_app a b) (texp_app a' b).
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSApp2 : forall a b b',
  mtstep b b' -> tval a -> mtstep (texp_app a b) (texp_app a b').
Proof.
  induction 1; intros; simpl; eauto.
Qed.

Theorem MSSucc : forall a a',
  mtstep a a' -> mtstep (texp_succ a) (texp_succ a').
Proof. induction 1; intros; simpl; eauto. Qed.

Theorem MSPair2 : forall a b b',
     mtstep b b' -> tval a
  -> mtstep (texp_pair a b) (texp_pair a b').
Proof. induction 1; intros; eauto. Qed.

Theorem MSPair : forall a a' b b',
     mtstep a a' -> mtstep  b b' -> tval a'
  -> mtstep (texp_pair a b) (texp_pair a' b').
Proof.
  induction 1; intros; [ eauto using MSPair2
                       | eapply MSTStep; eauto ].
Qed.

Theorem MSNatCase1 : forall a a' b1 b2,
  mtstep a a' -> mtstep (texp_ncase a b1 b2) (texp_ncase a' b1 b2).
Proof.
  induction 1; eauto.
Qed.


Theorem MSPCase1 : forall a a' b,
  mtstep a a' -> mtstep (texp_pcase a b) (texp_pcase a' b).
Proof.
  induction 1; eauto.
Qed.

Theorem open_ptstep_into : forall A a a' m,
  ptstep a a' -> ptstep ({m ~t> a} A) ({m ~t> a'}A).
Proof.
 induction A; simpl; intros; auto; eauto.
   db_var_case.
Qed.

Theorem open_mptstep_into : forall A a a' m ,
  mptstep a a' -> mptstep ({m ~t> a} A) ({m ~t> a'}A).
Proof.
  induction 1; eauto using open_ptstep_into.
Qed.  

Theorem mptstep_zero_inversion' : forall A B,
  mptstep A B -> A = texp_zero -> B = texp_zero.
Proof.
  induction 1; intros; subst; auto;
  match goal with
  | [ H : ptstep texp_zero _ |- _ ] => inversion H; subst; eauto
  end.
Qed.

Theorem mptstep_zero_inversion : forall A,
  mptstep texp_zero A -> A = texp_zero.
Proof.  intros; eapply mptstep_zero_inversion'; eauto.  Qed.

Theorem mptstep_succ_inversion' : forall A B,
     mptstep A B -> forall A1, A = texp_succ A1
  -> exists B1,
       B = texp_succ B1 /\ mptstep A1 B1.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : ptstep (texp_succ _) _ |- _ ] =>
    inversion H; subst; destruct_ih; subst; eauto 8
  end.
Qed.

Theorem mptstep_succ_inversion : forall A B,
     mptstep (texp_succ A) B
  -> exists B1,
       B = texp_succ B1 /\ mptstep A B1.
Proof. eauto using mptstep_succ_inversion'. Qed.

Ltac mpstep_inversion :=
  repeat (match goal with
  | [ H : mptstep texp_zero ?B |- _ ] => let H' := fresh H in
    match B with
    | texp_zero => fail 1
    | _ => assert (H' := mptstep_zero_inversion _ H); subst; try solve [inversion H']
    end
  | [ H : mptstep (texp_succ ?A') ?B |- _ ] => let H' := fresh H in
    match B with 
    | texp_succ ?B' => 
      match goal with
      | [ H : mptstep A' B' |- _ ] => fail 2
      | _ => idtac
      end
    | _ =>
      destruct (mptstep_succ_inversion _ _ H) as [? [HEq ?]]; subst; 
      try solve [inversion HEq]
    end
  end). 
